from .bigint import BigInt
from .binary import Binary
from .bit import Bit
from .char import Char
from .date import Date
from .decimal import Decimal
from .double import Double
from .float import Float
from .integer import Integer
from .longnvarchar import LongNVarChar
from .longvarbinary import LongVarBinary
from .longvarchar import LongVarChar
from .nchar import NChar
from .numeric import Numeric
from .nvarchar import NVarChar
from .real import Real
from .smallint import SmallInt
from .time import Time
from .timestamp import Timestamp
from .tinyint import TinyInt
from .unknown import Unknown
from .varbinary import VarBinary
from .varchar import VarChar

__all__ = [
    "BigInt",
    "Binary",
    "Bit",
    "Char",
    "Date",
    "Decimal",
    "Double",
    "Float",
    "Integer",
    "LongVarBinary",
    "LongVarChar",
    "Numeric",
    "Real",
    "SmallInt",
    "Time",
    "Timestamp",
    "TinyInt",
    "Unknown",
    "VarBinary",
    "VarChar",
    "NChar",
    "LongNVarChar",
    "NVarChar",
]



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing the Environment API client."""

import requests
from ibm_watsonx_data_integration.common.constants import PROD_BASE_URL
from ibm_watsonx_data_integration.common.utils import wait_and_retry_on_http_error
from ibm_watsonx_data_integration.cpd_api.base import BaseAPIClient
from typing import TYPE_CHECKING
from urllib.parse import quote

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator

DEFAULT_ENVIRONMENT_API_VERSION = 1  # Adjust the version number as needed


# Assisted by watsonx Code Assistant
class EnvironmentApiClient(BaseAPIClient):
    """The API client of the Environment service."""

    def __init__(
        self,
        auth: "BaseAuthenticator",  # pragma: allowlist secret
        base_url: str = PROD_BASE_URL,
    ) -> None:
        """The __init__ of the EnvironmentApiClient.

        Args:
            auth: The Authentication object.
            base_url: The Cloud Pak for Data URL. Default: "https://api.dataplatform.cloud.ibm.com"
        """
        super().__init__(auth=auth, base_url=base_url)
        self.url_path_sset_env = f"sset/engine_manager/v{DEFAULT_ENVIRONMENT_API_VERSION}/streamsets_environments"
        self.url_path_sset_engine_versions = (
            f"sset/engine_manager/v{DEFAULT_ENVIRONMENT_API_VERSION}/streamsets_engine_versions"
        )

    def get_environment(self, environment_id: str, params: dict) -> requests.Response:
        """Get a StreamSets Environment by ID.

        Args:
            environment_id: Environment id.
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}/{quote(environment_id, safe='')}"
        response = self.get(url=url, params=params)
        return response

    def delete_environment(self, environment_id: str, params: dict) -> requests.Response:
        """Deletes an environment by its ID.

        Args:
            environment_id: Environment id.
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}/{quote(environment_id, safe='')}"
        response = self.delete(url=url, params=params)
        return response

    def patch_environment(self, data: str, environment_id: str, params: dict) -> requests.Response:
        """Updates an existing environment with patch option.

        Args:
            data: List of patch operation to do on an environment.
            environment_id: Environment id.
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}/{quote(environment_id, safe='')}"
        response = self.patch(
            url=url, data=data, params=params, headers={"Content-Type": "application/json-patch+json"}
        )
        return response

    @wait_and_retry_on_http_error()
    def get_environments(self, params: dict) -> requests.Response:
        """Retrieves a list of all environments.

        Args:
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}"
        response = self.get(url=url, params=params)
        return response

    def create_environment(self, data: str, params: dict) -> requests.Response:
        """Creates a new environment.

        Args:
            data: Payload required to create new Environment.
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}"
        response = self.post(url=url, data=data, params=params)
        return response

    def get_docker_run_command(self, environment_id: str, params: dict) -> requests.Response:
        """Get the Docker run command for a StreamSets Engine in a StreamSets Environment.

        Args:
            environment_id: Environment id.
            params: Query parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}/{quote(environment_id, safe='')}/engine_docker_run_command"
        response = self.get(url=url, params=params, headers={"Accept": "text/plain"})
        return response

    def get_engines(self, environment_id: str, params: dict) -> requests.Response:
        """Lists all StreamSets Engines in a StreamSets Environment.

        Args:
            environment_id: Environment id.
            params: Query parameters.

        Returns:
            Response from the API.
        """
        url = f"{self.base_url}/{self.url_path_sset_env}/{quote(environment_id, safe='')}/engines"
        response = self.get(url=url, params=params)
        return response

    def get_engine_versions(self) -> requests.Response:
        """List all StreamSets Engine Versions.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_engine_versions}"
        response = self.get(url=url)
        return response

    def get_engine_by_version(self, engine_version: str) -> requests.Response:
        """Get a StreamSets Engine info by via version name.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_engine_versions}/{quote(engine_version, safe='')}"
        response = self.get(url=url)
        return response

    def get_library_definitions_for_engine_version(self, engine_version: str) -> requests.Response:
        """Get a library definitions of a particular engine version, includes all possible stage libraries.

        Args:
            engine_version: Version of the engine for which the library definitions should be fetched.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_sset_engine_versions}/{quote(engine_version)}/definitions"
        return self.get(url=url)



from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from pydantic import ConfigDict, Field
from typing import ClassVar


class Match360Conn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "99265578-2e54-4b6b-baea-3058fc2ecc96"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    ds_host: str | None = Field(None, alias="_host")
    ds_port: str | None = Field(None, alias="_port")
    api_key: str = Field(None, alias="api_key")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    cpd_user: str | None = Field(None, alias="cpd_user")
    crn: str | None = Field(None, alias="crn")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    gateway_url: str | None = Field(None, alias="gateway_url")
    iam_url: str | None = Field(None, alias="iam_url")
    mdm_instance_id: str | None = Field(None, alias="mdm_instance_id")
    route_host: str | None = Field(None, alias="route_host")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        include.add("api_key") if (not self.defer_credentials) else exclude.add("api_key")
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")
        include.add("mdm_instance_id") if (self.hidden_dummy_property1) else exclude.add("mdm_instance_id")
        include.add("cpd_user") if (self.hidden_dummy_property1) else exclude.add("cpd_user")
        include.add("route_host") if (self.hidden_dummy_property1) else exclude.add("route_host")

        include.add("api_key") if (not self.defer_credentials) else exclude.add("api_key")
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing the Job API client."""

import requests
from ibm_watsonx_data_integration.cpd_api.base import BaseAPIClient
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator

DEFAULT_JOBS_COMMON_CORE_API_VERSION = 2


class JobApiClient(BaseAPIClient):
    """The API client for resources related with Jobs and Job Runs."""

    def __init__(
        self,
        auth: "BaseAuthenticator",  # pragma: allowlist secret
        base_url: str = "https://api.dataplatform.cloud.ibm.com",
    ) -> None:
        """The __init__ of the JobApiClient.

        Args:
            auth: The Authentication object.
            base_url: The Cloud Pak for Data URL.
        """
        super().__init__(auth=auth, base_url=base_url)
        self.url_path_common_core = f"v{DEFAULT_JOBS_COMMON_CORE_API_VERSION}/jobs"

    def get_jobs(self, params: dict[str, Any]) -> requests.Response:
        """Retrieves list of Jobs for specified project.

        Args:
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}"
        response = self.get(url=url, params=params)
        return response

    def get_job(self, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Retrieves information about selected Job.

        Args:
            job_id: The ID of the job to use.
            params: Query Parameters.
        f
        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}"
        response = self.get(url, params=params)
        return response

    def create_job(self, data: dict, params: dict[str, Any]) -> requests.Response:
        """Creates a Job instance.

        Args:
            data: Payload required to create a new Job.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}"
        response = self.post(url, data=data, params=params)
        return response

    def delete_job(self, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Remove a Job.

        Args:
            job_id: The ID of the job to use.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}"
        response = self.delete(url, params=params)
        return response

    def update_job(self, job_id: str, data: list[dict], params: dict[str, Any]) -> requests.Response:
        """Patch Job attributes.

        Args:
            job_id: The ID of the job to use.
            data: List of patch operation to do on a Job.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}"
        response = self.patch(url, data=data, params=params)
        return response

    def get_job_runs(self, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Retrieves list of Job Runs for a Job.

        Args:
            job_id: The ID of the job to use.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs"
        response = self.get(url, params=params)
        return response

    def get_job_run(self, run_id: str, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Retrieves information about selected Job Run.

        Args:
            run_id: The ID of the job run.
            job_id: The ID of the job to use.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs/{quote(run_id, safe='')}"
        response = self.get(url, params=params)
        return response

    def delete_job_run(self, run_id: str, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Delete single run for a Job.

        Args:
            run_id: The ID of the job run.
            job_id: The ID of the job to use.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs/{quote(run_id, safe='')}"
        response = self.delete(url, params=params)
        return response

    def create_job_run(self, job_id: str, data: dict, params: dict[str, Any]) -> requests.Response:
        """Create Job Run instance for given asset.

        Args:
            job_id: The ID of the job to use.
            data: Payload with job run configuration.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs"
        response = self.post(url, data=data, params=params)
        return response

    def cancel_job_run(self, run_id: str, job_id: str, data: dict, params: dict[str, Any]) -> requests.Response:
        """Cancel already running job run.

        Args:
            run_id: The ID of the job run.
            job_id: The ID of the job to use.
            data: Required by POST endpoint, actually sends empty dictionary.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = (
            f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs/{quote(run_id, safe='')}/cancel"
        )
        response = self.post(url, data=data, params=params)
        return response

    def get_job_run_logs(self, run_id: str, job_id: str, params: dict[str, Any]) -> requests.Response:
        """Retrieves list of logs for job run.

        Args:
            run_id: The ID of the job run.
            job_id: The ID of the job to use.
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/{quote(job_id, safe='')}/runs/{quote(run_id, safe='')}/logs"
        response = self.get(url, params=params)
        return response

    def get_swagger(self) -> requests.Response:
        """Retrieve the swagger definitions to retrieve projects.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path_common_core}/docs/schemas/v2-swagger.yaml"
        response = self.get(url=url)
        return response



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Modules containing UserProfile Model and UserSettings Model."""

from __future__ import annotations

import json
import requests
from enum import Enum
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from pydantic import ConfigDict, Field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.platform import Platform


class UserState(Enum):
    """An enumeration representing the possible states of a user's account."""

    PROCESSING = "PROCESSING"
    PENDING = "PENDING"
    ACTIVE = "ACTIVE"
    DISABLED_CLASSIC_INFRASTRUCTURE = "DISABLED_CLASSIC_INFRASTRUCTURE"
    VPN_ONLY = "VPN_ONLY"


class UserProfile(BaseModel):
    """Model representing a user profile.

    Attributes:
        id: Unique identifier of the user.
        iam_id: The user's IAM ID.
        realm: The realm to which the user belongs.
        user_id: The user's identifier (often their email address).
        first_name: The first name (given name) of the user.
        last_name: The last name (family name) of the user.
        state: The current state of the user's account (e.g. PROCESSING, PENDING, ACTIVE,
            DISABLED_CLASSIC_INFRASTRUCTURE, VPN_ONLY).
        sub_state: The substate of the user's account, if applicable.
        email: The email address of the user.
        phone: The primary phone number of the user.
        alt_phone: An alternative phone number for the user.
        photo: A URL or reference to the user's photo.
        account_id: Unique identifier of the account the user belongs to.
        added_on: Timestamp of when the user was added.
        invited_on: Timestamp of when the user was invited.
    """

    id: str = Field(repr=True)
    iam_id: str = Field(repr=True)
    realm: str = Field(repr=False)
    user_id: str = Field(repr=True)
    first_name: str | None = Field(default=None, alias="firstname", repr=False)
    last_name: str | None = Field(default=None, alias="lastname", repr=False)
    state: UserState = Field(repr=True)
    sub_state: str | None = Field(default=None, repr=False)
    email: str = Field(repr=True)
    phone: str | None = Field(default=None, alias="phonenumber", repr=False)
    alt_phone: str | None = Field(default=None, alias="altphonenumber", repr=False)
    photo: str | None = Field(default=None, repr=False)
    account_id: str = Field(repr=True)
    added_on: str = Field(repr=True)
    invited_on: str | None = Field(default=None, alias="invitedOn", repr=False)

    model_config = ConfigDict(use_enum_values=True, frozen=True)
    _expose: bool = True

    def __init__(self, platform: Platform | None = None, **user_profile_json: dict) -> None:
        """The __init__ of the User Profile.

        Args:
            platform: The Platform object.
            user_profile_json: The JSON for the User Profile.
        """
        super().__init__(**user_profile_json)
        self._platform = platform
        if platform is not None and hasattr(platform, "_user_api"):
            self._user_api = platform._user_api

    @property
    def settings(self) -> UserSettings:
        """Returns the user settings associated with the current IAM identity and the current account.

        Returns:
            A User Settings instance retrieved from the API.
        """
        resp = self._user_api.get_user_settings(self.account_id, self.iam_id)
        return UserSettings(platform=self._platform, **resp.json())

    def update_user_settings(self, user_settings: UserSettings) -> requests.Response:
        """Update the settings for a specific user in an account.

        Args:
            user_settings: Instance of a UserSettings to update.

        Returns:
            A HTTP response.
        """
        payload = {
            "language": user_settings.language,
            "notification_language": user_settings.notification_language,
            "allowed_ip_addresses": user_settings.allowed_ip_addresses,
            "self_manage": user_settings.self_manage,
        }
        return self._user_api.update_user_settings(
            account_id=self.account_id, iam_id=self.iam_id, data=json.dumps(payload)
        )

    @property
    def type(self) -> str:
        """This property returns the member type "user".

        Returns:
            The member type.
        """
        return "user"


class UserSettings(BaseModel):
    """Model representing user settings preferences.

    Attributes:
        language: The user's preferred language. This field may be empty.
        notification_language: The preferred language for notifications. This field may be empty.
        allowed_ip_addresses: A string representing allowed IP addresses (e.g., comma-separated).
            This field may be empty.
        self_manage: Flag indicating if the user has permission to manage their settings.
        two_fa: Flag that indicates whether two-factor authentication is enabled.
                This field is mapped from the JSON key "2FA".
        security_questions_required: Flag indicating if security questions are required for the user.
        security_questions_setup: Flag that indicates whether the user has set up their security questions.
    """

    language: str = Field(default="", repr=False)
    notification_language: str = Field(default="", repr=False)
    allowed_ip_addresses: str = Field(default="", repr=False)
    self_manage: bool = Field(default=False, repr=True)
    two_fa: bool = Field(default=False, alias="2FA", repr=False)
    security_questions_required: bool = Field(default=False, repr=False)
    security_questions_setup: bool = Field(default=False, repr=False)

    _expose: bool = True

    def __init__(self, platform: Platform | None = None, **settings_json: dict) -> None:
        """The __init__ of the UserSettings model.

        Args:
            platform: The Platform object.
            settings_json: The JSON for the user settings.
        """
        super().__init__(**settings_json)
        self._platform = platform


class UserProfiles(CollectionModel):
    """Collection of UserProfile instances."""

    def __init__(self, platform: Platform) -> None:
        """The __init__ of the UserProfiles class.

        Args:
            platform: The Platform object.
        """
        super().__init__(platform)
        self.unique_id = "id"
        self.account_id = platform.current_account.account_id

    def __len__(self) -> int:
        """Total number of user profiles."""
        query_params = {
            "limit": 1,
        }
        res = self._platform._user_api.get_users(self.account_id, params=query_params)
        res_json = res.json()
        return res_json["total_results"]

    def _request_parameters(self) -> list:
        return ["iam_id", "limit", "_start", "user_id"]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        if "_start" in request_params:
            request_params["_start"] = request_params["_start"].split("_start=")[1].split("&")[0]

        request_params_defaults = {
            "iam_id": None,
            "limit": None,
            "_start": None,
            "user_id": None,
        }
        request_params_unioned = request_params_defaults.copy()
        request_params_unioned.update(request_params)

        iam_id = request_params_unioned.get("iam_id")
        user_id = request_params_unioned.get("user_id")

        if iam_id and user_id:
            raise ValueError("Only one of 'iam_id' or 'user_id' can be provided, not both.")

        if iam_id:
            response = self._platform._user_api.get_user_profile(self.account_id, iam_id).json()
            response = {"resources": [response]}
        else:
            response = self._platform._user_api.get_users(
                account_id=self.account_id,
                params={k: v for k, v in request_params_unioned.items() if v is not None},
            ).json()
        return CollectionModelResults(
            response,
            UserProfile,
            "next_url",
            "_start",
            "resources",
            {"platform": self._platform},
        )



from ibm_watsonx_data_integration.services.datastage.models.flow.datastage_flow import DataStageFlow
from ibm_watsonx_data_integration.services.datastage.tests.integration.util import AbstractFlowRunTest
from os import environ
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestDb2BigSqlConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "db2BigSql_peek"

    def get_fc(self, project: Project):
        fc = DataStageFlow(project=project, name=self.get_flow_name(), description="", flow_type="datastage")

        # Stages
        _test_before_after_tmp_1_1_ = fc.add_stage("IBM Db2 Big SQL", "TEST_BEFORE_AFTER_TMP_1_1")
        _test_before_after_tmp_1_1_.configuration.schema_name = "BIGSQL"
        _test_before_after_tmp_1_1_.configuration.table_name = "TEST_BEFORE_AFTER_TMP_1"

        _test_before_after_tmp_1_1_.configuration.connection.database = "BIGSQL"
        _test_before_after_tmp_1_1_.configuration.connection.hostname_or_ip_address = environ["DB2BIGSQL_HOST"]
        _test_before_after_tmp_1_1_.configuration.connection.password = environ["DB2BIGSQL_PASSWORD"]
        _test_before_after_tmp_1_1_.configuration.connection.port = "32051"
        _test_before_after_tmp_1_1_.configuration.connection.port_is_ssl_enabled = False
        _test_before_after_tmp_1_1_.configuration.connection.username = environ["DB2BIGSQL_USERNAME"]

        _peek_1_ = fc.add_stage("Peek", "Peek_1")

        # Graph
        test_before_after_tmp_1_1_schema = (
            _test_before_after_tmp_1_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()
        )

        # Schemas
        test_before_after_tmp_1_1_schema.add_field("CHAR", "COL1").nullable().length(5)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



from ibm_watsonx_data_integration.services.datastage.models import flow_json_model as models
from ibm_watsonx_data_integration.services.datastage.models.enums import FIELD
from ibm_watsonx_data_integration.services.datastage.models.schema.field.base_field import BaseField
from typing import TypeVar
T = TypeVar("T", bound="LongVarBinary")

class LongVarBinary(BaseField):
    def __init__(self, name: str):
        super().__init__(name)
        self._set_model_type("binary")
        self._set_app_type_code("BINARY")
        self._set_app_odbc_type("LONGVARBINARY")
        self.configuration.metadata.min_length = 0
        self.configuration.metadata.max_length = 100

    def _create_from_model(self, model: models.FieldModelComplex)-> T:
        field = LongVarBinary(model.name)
        field.configuration = model
        return field

    def length(self, length: int):
        """Set the length of this field.

        Args:
            length: The length to set for this field.

        Returns:
            A new instance of the field with the updated length.

        """
        return self._length(length)._max_length(length)

    def byte_to_skip(self, num_bytes: int):
        return self._byte_to_skip(num_bytes)

    def delimiter(self, delim: FIELD.Delim):
        return self._delim(delim)

    def delimiter_string(self, delimiter_string: str):
        return self._delim_string(delimiter_string)

    def generate_on_output(self):
        return self._generate_on_output()

    def prefix_bytes(self, prefix_bytes: FIELD.Prefix):
        return self._prefix_bytes(prefix_bytes)

    def quote(self, quote_type: FIELD.Quote):
        return self._quote(quote_type)

    def start_position(self, position: int):
        return self._start_position(position)

    def tag_case_value(self, tag_case_value: int):
        return self._tagcase(tag_case_value)

    def link_field_reference(self, link_field_reference):
        return self._link_field_reference(link_field_reference)



# from ..components.local_message_handler import LocalMessageHandler
# from ..components.message_handler import MessageHandler
import requests
from ibm_watsonx_data_integration.common.models import CollectionModel, CollectionModelResults
from ibm_watsonx_data_integration.services.datastage.models.flow_stages import FlowComposer
from ibm_watsonx_data_integration.cpd_models.flow_model import Flow, PayloadExtender
from typing_extensions import override
from ibm_watsonx_data_integration.services.datastage.models.flow.dag import (
    DAG,
    Link,
    Node,
)
from ibm_watsonx_data_integration.services.streamsets.models.environment_model import Environment
from urllib.parse import parse_qs, urlparse

# from ibm_watsonx_data_integration.services.datastage._console import console
# from ibm_watsonx_data_integration.services.datastage.models.flow.dag import SuperNode
# from ibm_watsonx_data_integration.services.datastage.models.flow_runner import FlowRunner
from ibm_watsonx_data_integration.services.datastage.models.layout import LayeredLayout

# from ibm.datastage._framework.paramsets import LocalParameters, ParameterSet
# from ibm.datastage._framework.runtime import Runtime
# from ibm.datastage._framework.schema.data_definition import DataDefinition

import json

from typing import TYPE_CHECKING, Any
if TYPE_CHECKING:
    from ibm_watsonx_data_integration.cpd_models.project_model import Project
from ibm_watsonx_data_integration.services.datastage.models.extractor import FlowExtractor

@Flow.register("datastage")
class DataStageFlow(Flow, FlowComposer):
    rcp: bool = False
    acp: bool = True
    _project : "Project" = None
    name : str = "unnamed_flow"
    description : str = ""
    # env : Environment = None
    flow_id : str = ""

    def __init__(self,
                 project = None,
                 **kwargs):
        """Initialize a DataStage flow.

        Args:
           project: The project for the flow to be created in. If None, then the flow is not created remotely.
           name: The flow name.
           description: The flow description.
           flow_type: The type of flow. This should always be datastage.
        """

        Flow.__init__(self, **kwargs)
        FlowComposer.__init__(self, DAG())
        self._project = project

    # def use_data_definition(self, data_definition: DataDefinition):
    #     self.data_definitions.append(data_definition)
    #     return self

    # def use_message_handler(self, message_handler: MessageHandler):
    #     self.message_handlers.append(message_handler)
    #     return self

    # def use_local_message_handler(self, local_message_handler: LocalMessageHandler):
    #     self.local_message_handler = local_message_handler
    #     return self

    # def use_paramset(self, paramset: ParameterSet):
    #     self.parameter_sets.append(paramset)
    #     return self

    # def use_localparams(self, localparams: LocalParameters):
    #     self.local_parameters = localparams
    #     return self

    # def use_runtime(self, runtime: Runtime):
    #     self.runtime = runtime
    #     return self

    def use_runtime_column_propagation(self, rcp: bool = True):
        self.rcp = rcp
        return self

    def use_auto_column_propagation(self, acp: bool = True):
        self.acp = acp
        return self

    # def _add_ghost_node(self) -> Node:
    #     node = GhostNode(self._dag)
    #     self._dag.add_node(node)
    #     return node

    def remove_stage(self, node: Node) -> None:
        """Removes a node from the flow.

        Args:
            node: The node to be removed.
        """
        self._dag.remove_node(node)

    # def add_markdown_comment(self, text: str) -> MarkdownComment:
    #     comment = MarkdownComment(self._dag, content=text)
    #     self._dag.add_node(comment)
    #     return comment

    # def add_styled_comment(self, text: str) -> StyledComment:
    #     comment = StyledComment(self._dag, content=text)
    #     self._dag.add_node(comment)
    #     return comment

    def get_link(self, source: Node, destination: Node) -> Link:
        """Gets the link between two nodes in the flow. If there are no links or multiple links, it raises an error.

        Args:
            source: Node from which the link originates.
            destination: Node to which the link points.

        Returns:
            The single link between the source and destination nodes.
        """
        links = self._dag.get_links_between(source, destination)
        if len(links) == 0:
            raise ValueError("No link between nodes")
        if len(links) > 1:
            raise ValueError("Multiple links between nodes")
        return links[0]

    def get_links(self, source: Node, destination: Node) -> list[Link]:
        """Gets all links between two nodes in the flow.

        Args:
            source: Node from which the links originate.
            destination: Node to which the links point.

        Returns:
            A list of Links between the source and destination nodes.
        """
        return self._dag.get_links_between(source, destination)

    def _dag_to_json(self) -> str:
        self._dag.compute_metadata()
        lay = LayeredLayout(self._dag)
        lay.compute()

        # # Compute layout for child DAGs for local subflows
        # for node in self._dag.nodes():
        #     if isinstance(node, SuperNode) and node.is_local:
        #         node.subflow_dag.compute_metadata()
        #         sub_lay = LayeredLayout(node.subflow_dag)
        #         sub_lay.compute()

        ser = FlowExtractor(self)
        ser.extract()

        flow_model = ser.serialize()
        return flow_model.model_dump_json(indent=2, exclude_none=True, by_alias=True, warnings=False)

    @staticmethod
    def _create(project: "Project" = None,
                name: str = "unnamed_flow",
                environment: Environment = None,
                description: str = "",
                flow_type: str = "datastage",
                ) -> "DataStageFlow":
        # self.parameter_sets: list[ParameterSet] = []
        # self.local_parameters: LocalParameters | None = None
        # self.runtime: Runtime | None = None
        # self.message_handlers: list[MessageHandler] = []
        # self.local_message_handler: LocalMessageHandler | None = None
        # self.data_definitions: list[DataDefinition] = []

        flow_name = name

        test_json : dict = {
            "primary_pipeline" : "",
            "pipelines" : [ {
                "nodes" : [ ],
                "description" : "",
                "id" : "",
                "app_data" : {
                "datastage" : { },
                "ui_data" : {
                    "comments" : [ ]
                }
                },
                "runtime_ref" : ""
            } ],
            "json_schema" : "http://api.dataplatform.ibm.com/schemas/common-pipeline/pipeline-flow/pipeline-flow-v3-schema.json",
            "schemas" : [ ],
            "doc_type" : "pipeline",
            "id" : "",
            "app_data" : {
                "datastage" : { }
            },
            "version" : "3.0"
        }
        project_id = project.metadata.guid

        response = project._platform._datastage_flow_api.create_datastage_flows(
            data_intg_flow_name=flow_name,
            pipeline_flows=test_json,
            project_id=project_id,
        )

        flow = DataStageFlow(project=project, name=name, description=description, flow_type=flow_type, flow_id = response.json()["metadata"]["asset_id"])
        return flow

    @staticmethod
    def create_or_get(project: "Project" = None,
                name: str = "unnamed_flow",
                environment: Environment = None,
                description: str = "",
                flow_type: str = "datastage",
                ) -> "DataStageFlow":
        """Either creates the flow with the given name or returns the preexisting flow of that name."""
        try:
            return project.datastage_flows.get(name = name)
        except ValueError:
            return DataStageFlow._create(project, name, environment, description, flow_type)

    def _update(self) -> requests.Response:
        flow_name = self.name
        flow_json = json.loads(self._dag_to_json())
        project_id = self._project.metadata.guid
        response = self._project._platform._datastage_flow_api.update_datastage_flows(
            data_intg_flow_id=self.flow_id,
            data_intg_flow_name=flow_name,
            pipeline_flows=flow_json,
            project_id=project_id,
        )
        return response

    def _delete(self) -> requests.Response:
        return self._project._platform._datastage_flow_api.delete_datastage_flows(
            id=[self.flow_id],
            project_id=self._project.metadata.guid,
        )

    def compile(self) -> requests.Response:
        flow_json = self._dag_to_json()
        project_id = self._project.metadata.guid

        response = self._project._platform._datastage_flow_api.compile_datastage_flows(
            data_intg_flow_id=self.flow_id,
            pipeline_flows=flow_json,
            project_id=project_id,
        )
        return response

    def _duplicate(self, name : str, description: str) -> "DataStageFlow":
        project_id = self._project.metadata.guid

        response = self._project._platform._datastage_flow_api.clone_datastage_flows(
            data_intg_flow_id=self.flow_id,
            project_id=project_id,
            data_intg_flow_name=name,
            description=description
        )
        flow = DataStageFlow(project=self._project, name=name, description=description, flow_type="datastage", flow_id = response.json()["metadata"]["asset_id"])
        return flow

    def get_compile_status(self) -> requests.Response:
        project_id = self._project.metadata.guid
        response = self._project._platform._datastage_flow_api.get_flow_compile_status(data_intg_flow_id=self.flow_id, project_id=project_id)
        return response

    def get_compile_info(self) -> requests.Response:
        project_id = self._project.metadata.guid
        response = self._project._platform._datastage_flow_api.datastage_flows_compile_info(data_intg_flow_id=self.flow_id, project_id=project_id)
        return response

class DataStageFlows(CollectionModel):
    """Collection of DataStageFlow objects."""

    def __init__(self, project: "Project") -> None:
        """The __init__ of the DataStageFlows class.

        Args:
            project: The Project object.
        """
        super().__init__(project)
        self._project = project
        self.unique_id = "flow_id"

    def _request_parameters(self) -> list:
        return ["data_intg_flow_id",
                "project_id",
                "catalog_id",
                "space_id",
                "sort",
                "start",
                "limit",
                "entity.name",
                "entity.description"]

    def __len__(self) -> int:
        """The len of the DataStageFlows class."""
        query_params = {
            "project_id": self._project.metadata.guid,
        }
        res = self._project._platform._datastage_flow_api.list_datastage_flows(**query_params)
        res_json = res.json()
        return res_json["total_count"]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        if "start" in request_params:
            parsed_url = urlparse(request_params["start"]["href"])
            params = parse_qs(parsed_url.query)
            request_params["start"] = params.get("start", [None])[0]

        request_params_defaults = {
            "data_intg_flow_id": None,
            "project_id": self._project.metadata.guid,
            "catalog_id": None,
            "space_id": None,
            "sort": None,
            "start": None,
            "limit": 100,
            "entity.name": None,
            "entity.description": None,
        }
        request_params_unioned = request_params_defaults
        request_params_unioned.update(request_params)

        if "entity.name" in request_params_unioned:
            request_params_unioned["entity_name"] = request_params_unioned.get("entity.name")
        if "entity.description" in request_params_unioned:
            request_params_unioned["entity_description"] = request_params_unioned.get("entity.description")

        if "data_intg_flow_id" in request_params:
            response_json = self._project._platform._datastage_flow_api.get_datastage_flows(
                **{k: v for k, v in request_params_unioned.items() if v is not None},
                data_intg_flow_id=request_params["data_intg_flow_id"],
            ).json()
            response = {"data_flows": [response_json]}

        else:
            response = self._project._platform._datastage_flow_api.list_datastage_flows(
                **{k: v for k, v in request_params_unioned.items() if v is not None}
            ).json()

        # Create parameters for construction
        for flow_json in response["data_flows"]:
            flow_json["flow_id"] = flow_json["metadata"]["asset_id"]
            flow_json["name"] = flow_json["metadata"]["name"]
            flow_json["description"] = flow_json["metadata"]["description"]

        return CollectionModelResults(
            results=response,
            class_type=DataStageFlow,
            response_bookmark="next",
            request_bookmark="start",
            response_location="data_flows",
            constructor_params={"project": self._project},
        )

class DataStageFlowPayloadExtender(PayloadExtender):
    """DataStage flow extender setup also compiles the flow.

    :meta: private
    """

    @override
    def extend(self, payload: dict[str, Any], flow: Flow) -> dict[str, Any]:
        flow.compile()
        payload["asset_ref"] = flow.flow_id
        return payload



from ibm_watsonx_data_integration.services.datastage.models.enums import IBM_MQ, IBM_MQ_CONNECTION
from os import environ
from tests.integration.datastage.util import AbstractFlowRunTest
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestIbmMqConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "mq_peek"

    def get_fc(self, project: Project):
        fc = project.create_flow(name=self.get_flow_name(), environment=None, description="", flow_type="datastage")

        # Stages
        _mq_1_ = fc.add_stage("IBM MQ", "MQ_1")
        _mq_1_.configuration.message_quantity = 10
        _mq_1_.configuration.message_read_mode = IBM_MQ.MessageReadMode.delete
        _mq_1_.configuration.process_end_of_data_message = True
        _mq_1_.configuration.queue_name = "MQAUTH.QUEUE4"
        _mq_1_.configuration.wait_time = 10

        _mq_1_.configuration.connection.cas_lite_service_authorization_header = ""
        _mq_1_.configuration.connection.channel_name = "MQAUTH.CHAN"
        _mq_1_.configuration.connection.connection_name = "c-01.private.us-south.link.satellite.cloud.ibm.com(33427)"
        _mq_1_.configuration.connection.transport_type = IBM_MQ_CONNECTION.ClientChannelDefinitionTransportType.tcp
        _mq_1_.configuration.connection.defer_credentials = False
        _mq_1_.configuration.connection.password = environ["IBMMQ_PASSWORD"]
        _mq_1_.configuration.connection.queue_manager_name = "MQAUTH"
        _mq_1_.configuration.connection.ssl_connection = False
        _mq_1_.configuration.connection.username = environ["IBMMQ_USERNAME"]

        _peek_1_ = fc.add_stage("Peek", "Peek_1")
        _peek_1_.configuration.runtime_column_propagation = 0

        # Graph
        mq_1_schema = _mq_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        mq_1_schema.add_field("CHAR", "COLUMN_1").length(100)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Service Models."""

import json
import requests
from ibm_watsonx_data_integration.common.constants import (
    DATASTAGE,
    DEFAULT_RESOURCE_REGION_ID_MAP,
    RESOURCE_PLAN_ID_MAP,
)
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from pydantic import Field
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.platform import Platform


class Service(BaseModel):
    """The Model for Services."""

    id: str = Field(repr=False, frozen=True)
    guid: str = Field(repr=True, frozen=True)
    url: str = Field(repr=False, frozen=True)
    created_at: str = Field(repr=False, frozen=True)
    updated_at: str = Field(repr=False, frozen=True)
    deleted_at: str | None = Field(repr=False, frozen=True)
    created_by: str = Field(repr=False, frozen=True)
    updated_by: str = Field(repr=False, frozen=True)
    deleted_by: str = Field(repr=False, frozen=True)
    scheduled_reclaim_at: str | None = Field(repr=False, frozen=True)
    restored_at: str | None = Field(repr=False, frozen=True)
    scheduled_reclaim_by: str = Field(repr=False, frozen=True)
    restored_by: str = Field(repr=False, frozen=True)
    name: str = Field(repr=True)
    region_id: str = Field(repr=False, frozen=True)
    account_id: str = Field(repr=False, frozen=True)
    reseller_channel_id: str = Field(repr=False, frozen=True)
    resource_plan_id: str = Field(repr=False, frozen=True)
    resource_group_id: str = Field(repr=False, frozen=True)
    dashboard_url: None | str = Field(repr=False, default=None, frozen=True)
    resource_group_crn: str = Field(repr=False, frozen=True)
    target_crn: str = Field(repr=False, frozen=True)
    allow_cleanup: bool = Field(repr=False, frozen=True)
    crn: str = Field(repr=False, frozen=True)
    state: str = Field(repr=False, frozen=True)
    type: str = Field(repr=False, frozen=True)
    resource_id: str = Field(repr=False, frozen=True)
    last_operation: dict | None = Field(repr=False, frozen=True)
    resource_aliases_url: str = Field(repr=False, frozen=True)
    resource_bindings_url: str = Field(repr=False, frozen=True)
    resource_keys_url: str = Field(repr=False, frozen=True)
    plan_history: list | None = Field(repr=False, frozen=True)
    migrated: bool = Field(repr=False, frozen=True)
    controlled_by: str = Field(repr=False, frozen=True)
    locked: bool = Field(repr=False, frozen=True)
    onetime_credentials: bool = Field(repr=False, frozen=True)

    def __init__(self, platform: Optional["Platform"] = None, **service_json: dict) -> None:
        """The __init__ of the Service.

        Args:
            service_json: The JSON for the Service.
            platform: The Platform object.
        """
        super().__init__(**service_json)
        self._platform = platform

    @staticmethod
    def _create(
        platform: "Platform",
        instance_type: str,
        name: str,
        target: str | None = None,
        tags: list | None = None,
    ) -> "Service":
        resource_plan_id = RESOURCE_PLAN_ID_MAP[instance_type]
        resource_group_id = platform._resource_controller_api.get_resource_groups().json()["resources"][0]["id"]
        target = target if target else DEFAULT_RESOURCE_REGION_ID_MAP[instance_type]
        tags = tags if tags else []

        data = {
            "name": name,
            "target": target,
            "resource_plan_id": resource_plan_id,
            "resource_group": resource_group_id,
            "tags": tags,
        }

        response = platform._resource_controller_api.create_resource_instance(json.dumps(data))

        return Service(platform=platform, **response.json())

    def _delete(self, delete_keys: bool = True) -> requests.Response:
        return self._platform._resource_controller_api.delete_resource_instance(self.guid, recursive=delete_keys)


class Services(CollectionModel):
    """Collection of Service instances."""

    def __init__(self, platform: "Platform") -> None:
        """The __init__ of the Services class.

        Args:
            platform: The Platform object.
        """
        super().__init__(platform)
        self._platform = platform
        self.unique_id = "id"
        self.resource_id = next((k for k, v in platform._service_id_to_name_map.items() if v == DATASTAGE))

    def __len__(self) -> int:
        """The len of the Services class."""
        res = self._platform._resource_controller_api.get_resource_instances({"type": "service_instance"})
        res_json = res.json()
        return res_json["rows_count"]

    def _request_parameters(self) -> list:
        return ["start", "type", "limit"]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        if request_params.get("start", None):
            request_params["start"] = request_params["start"].split("start=")[1]

        request_params_defaults = {"start": None, "type": "service_instance"}
        request_params_unioned = request_params_defaults
        request_params_unioned.update(request_params)
        response = self._platform._resource_controller_api.get_resource_instances(
            params={k: v for k, v in request_params_unioned.items() if v is not None}
        ).json()
        return CollectionModelResults(
            response,
            Service,
            "next_url",
            "start",
            "resources",
            {"platform": self._platform},
        )



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing the Metering API client."""

import requests
from ibm_watsonx_data_integration.cpd_api.base import BaseAPIClient
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator

DEFAULT_METERING_API_VERSION = 1


class MeteringApiClient(BaseAPIClient):
    """The API client for resources related with Jobs and Job Runs."""

    def __init__(
        self,
        auth: "BaseAuthenticator",  # pragma: allowlist secret
        base_url: str = "https://api.dai.dataplatform.cloud.ibm.com",
    ) -> None:
        """The __init__ of the JobApiClient.

        Args:
            auth: The Authentication object.
            base_url: The Cloud Pak for Data URL.
        """
        super().__init__(auth=auth, base_url=base_url)
        self.url_path = f"sset/metering/v{DEFAULT_METERING_API_VERSION}"

    def get_billing_data(self, params: dict[str, Any]) -> requests.Response:
        """Retrieves metering billing data.

        Args:
            params: Query Parameters.
        f
        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/metering_data/billing"
        response = self.get(url, params=params)
        return response

    def get_engine_metrics(self, params: dict[str, Any]) -> requests.Response:
        """Retrieves meterrics based on engine id.

        Args:
            params: Query Parameters.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/metering_event/utilization_by_engine_id"
        response = self.get(url, params=params)
        return response

    def save_engine_metrics(
        self,
        data: dict,
    ) -> requests.Response:
        """Send a payload of metrics mostly from Engine Manager and metering will save it.

        Args:
            data: Requires by POST endpoint.

        Returns:
            A HTTP response.
        """
        url = f"{self.base_url}/{self.url_path}/metering_event/save"
        response = self.post(url, data=data)
        return response



import inspect
import pydantic
from abc import ABC
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from ibm_watsonx_data_integration.services.datastage.models.schema import Schema


class Node(ABC):
    def __init__(self, dag: "DAG"):
        self._dag = dag
        self.metadata: dict[str, Any] = {}

        try:
            self.var_name = inspect.stack()[-1].code_context[0].split("=")[0].strip()
        except:
            self.var_name = "node"

    def connect_output_to(self, node: "Node") -> "Link":
        """Connects the output of this node to the input of another node.

        Args:
            node: The destination node to connect to.

        Returns:
            A Link object representing the connection.
        """
        link = Link(self._dag)
        link.src = self
        link.dest = node
        self._dag.add_link(link)
        return link

    def connect_input_to(self, node: "Node") -> "Link":
        """Connects the input of this node to the output of another node.

        Args:
            node: The source node to connect from.

        Returns:
            A Link object representing the connection.
        """
        link = Link(self._dag)
        link.src = node
        link.dest = self
        self._dag.add_link(link)
        return link

    def disconnect_output_from(self, dest: "Node") -> list["Link"]:
        """Disconnects the output of this node from the input of another node.

        Args:
            dest: The destination node to disconnect from.

        Returns:
            The list of Links that were removed.
        """
        links = self._dag.remove_link(self, dest, self.get_link(dest).name)
        if not links:
            raise ValueError(f"No link found between {self} and {dest}")
        return links

    def disconnect_input_from(self, src: "Node") -> list["Link"]:
        """Disconnects the input of this node from the output of another node.

        Args:
            src: The source node to disconnect from.

        Returns:
            The list of Links that were removed.
        """
        links = self._dag.remove_link(src, self, self.get_link(src).name)
        if not links:
            raise ValueError(f"No link found between {src} and {self}")
        return links

    def get_link(self, dest: "Node") -> "Link":
        links = self.get_links_to(dest)
        if links == 0:
            raise ValueError(f"No links found between {self} and {dest}")
        if len(links) > 1:
            raise ValueError(f"Multiple links found between {self} and {dest}")
        return links[0]

    def get_links_to(self, dest: "Node") -> list["Link"]:
        """Returns:
        All links between this node and the destination node.
        """
        return self._dag.get_links_between(self, dest)

    def _get_min_primary_inputs(self) -> int:
        """Returns:
        The minimum number of primary inputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_max_primary_inputs(self) -> int:
        """Returns:
        The maximum number of primary inputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_min_primary_outputs(self) -> int:
        """Returns:
        The minimum number of primary outputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_max_primary_outputs(self) -> int:
        """Returns:
        The maximum number of primary outputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_min_reject_outputs(self) -> int:
        """Returns:
        The minimum number of reject outputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_max_reject_outputs(self) -> int:
        """Returns:
        The maximum number of reject outputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_min_reference_inputs(self) -> int:
        """Returns:
        The minimum number of reference inputs for this node. -1 indicates no limit.
        """
        return -1

    def _get_max_reference_inputs(self) -> int:
        """Returns:
        The maximum number of reference inputs for this node. -1 indicates no limit.
        """
        return -1


class Link:
    def __init__(self, dag: "DAG", name: str = None):
        self._dag = dag
        self.name = name
        self.type = "PRIMARY"
        self.schema: Schema | None = None
        self.maps_to_link: str | None = None
        self.maps_from_link: str | None = None
        self.__src = None
        self.__dest = None

    @property
    def src(self):
        return self.__src

    @src.setter
    def src(self, new_left: Node):
        self.__src = new_left

    @property
    def dest(self):
        return self.__dest

    @dest.setter
    def dest(self, new_right: Node):
        self.__dest = new_right

    def set_name(self, name: str) -> "Link":
        self.name = name
        return self

    def primary(self) -> "Link":
        """Sets the link type to primary. Modifies the current link object.

        Returns:
            The current link object.
        """
        self.type = "PRIMARY"
        return self

    def reference(self) -> "Link":
        """Sets the link type to reference. Modifies the current link object.

        Returns:
            The current link object.
        """
        self.type = "REFERENCE"
        return self

    def reject(self) -> "Link":
        """Sets the link type to reject. Modifies the current link object.

        Returns:
            The current link object.
        """
        self.type = "REJECT"
        return self

    def map_to_link(self, link_name: str) -> "Link":
        self.maps_to_link = link_name
        return self

    def map_from_link(self, link_name: str) -> "Link":
        self.maps_from_link = link_name
        return self

    def create_schema(self) -> "Schema":
        """Initializes the schema that the link uses.

        Returns:
            The new link object.
        """
        self.schema = Schema()
        return self.schema


class DAG:
    def __init__(self):
        self.adj: dict[Node, dict[Node, list[Link]]] = {}
        self.is_metadata_computed = False

    def stages(self):
        for stage in self.adj.keys():
            yield stage

    def nodes(self):
        for node in self.adj.keys():
            yield node

    def links(self):
        for from_node, to_dict in self.adj.items():
            for to_node, links in to_dict.items():
                yield from links

    def links_stable(self):
        top_order = self.get_topological_ordering(stages_only=True)
        for node in top_order:
            to_nodes = list(self.adj[node].keys())
            to_nodes.sort(key=lambda n: top_order.index(n))
            for to_node in to_nodes:
                yield from self.adj[node][to_node]

    def get_links_between(self, src: "Node", dest: "Node") -> list[Link]:
        return self.adj.get(src, {}).get(dest, [])

    def add_node(self, node: "Node") -> "Node":
        self.adj.setdefault(node, {})
        return node

    def replace_node(self, old_node: Node, new_node: Node):
        if old_node not in self.adj:
            raise ValueError(f"Node {old_node.label} not found in DAG")

        self.adj[new_node] = self.adj.pop(old_node)

        for from_node, to_dict in self.adj.items():
            if old_node in to_dict:
                to_dict[new_node] = to_dict.pop(old_node)

        for to_node, links in self.adj[new_node].items():
            for link in links:
                if link.src == old_node:
                    link.src = new_node
        for from_node, links in self.adj.items():
            for to_node, links in self.adj[from_node].items():
                for link in links:
                    if link.dest == old_node:
                        link.dest = new_node

        if hasattr(old_node, "metadata"):
            new_node.metadata = old_node.metadata.copy()
            new_node.metadata["var_name"] = old_node.var_name
        else:
            new_node.metadata = {"var_name": old_node.var_name}
        self.remove_node(old_node)

    def remove_node(self, node: "Node"):
        if node in self.adj:
            del self.adj[node]
        for from_node, to_dict in self.adj.items():
            if node in to_dict:
                del to_dict[node]

    def add_link(self, link: "Link"):
        from_node = self.add_node(link.src)
        to_node = self.add_node(link.dest)

        if link in self.adj.get(from_node, {}).get(to_node, []):
            return

        self.adj[from_node].setdefault(to_node, [])
        self.adj[from_node][to_node].append(link)

    def remove_link(self, src: "Node", dest: "Node", link_name: str) -> list[Link]:
        removed = []
        if src in self.adj and dest in self.adj[src]:
            kept = []
            for l in self.adj[src][dest]:
                if l.name == link_name:
                    removed.append(l)
                else:
                    kept.append(l)
            if not kept:
                del self.adj[src][dest]
            else:
                self.adj[src][dest] = kept
        return removed

    def remove_links(self, src: "Node", dest: "Node") -> list[Link]:
        if src in self.adj and dest in self.adj[src]:
            links = self.adj[src][dest]
            del self.adj[src][dest]
            return links
        return []

    def get_topological_ordering(self, *, stages_only: bool):
        if stages_only:
            in_degrees = self.__compute_in_degrees(lambda n: isinstance(n, StageNode))
        else:
            in_degrees = self.__compute_in_degrees()

        # Queue of nodes with in-degree 0
        queue: list[Node] = [node for node, in_degree in in_degrees.items() if in_degree == 0]

        # Perform topological sorting of nodes using Kahn's algorithm
        top_order = []
        while queue:
            node = queue.pop(0)
            top_order.append(node)
            for to_node, links in self.adj[node].items():
                in_degrees[to_node] -= len(links)
                if in_degrees[to_node] == 0:
                    queue.append(to_node)

        # Check if graph has cycle
        if len(top_order) != len(self.adj):
            raise ValueError(f"Graph contains cycle {self.__print_cycle(self.adj)}")

        return top_order

    def compute_metadata(self):
        in_degrees_stages = self.__compute_in_degrees(lambda n: isinstance(n, StageNode))
        out_degrees_stages = self.__compute_out_degrees(lambda n: isinstance(n, StageNode))

        for node in self.nodes():
            node.metadata["in_degree"] = in_degrees_stages[node]
            node.metadata["out_degree"] = out_degrees_stages[node]
            node.metadata["parents"] = []
            node.metadata["children"] = []

        top_order = self.get_topological_ordering(stages_only=False)

        # Check if graph has cycle
        if len(top_order) != len(self.adj):
            raise ValueError(f"Graph contains cycle {self.__print_cycle(self.adj)}")

        # Compute parents of each node
        for link in self.links():
            # If link src is already in dest's parents, skip (this can happen due to multi-links between 2 nodes)
            if link.src not in link.dest.metadata["parents"]:
                link.dest.metadata["parents"].append(link.src)
            # Do the same for src's children
            if link.dest not in link.src.metadata["children"]:
                link.src.metadata["children"].append(link.dest)

        # Compute cardinality
        for node in self.nodes():
            if not isinstance(node, StageNode):
                continue

            node.metadata["primary_inputs"] = 0
            node.metadata["primary_outputs"] = 0
            node.metadata["reference_inputs"] = 0
            node.metadata["reference_outputs"] = 0
            node.metadata["reject_inputs"] = 0
            node.metadata["reject_outputs"] = 0

            for parent in node.metadata["parents"]:
                if not isinstance(parent, StageNode):
                    continue

                links = self.adj[parent][node]
                for link in links:
                    match link.type:
                        case "PRIMARY":
                            node.metadata["primary_inputs"] += 1
                        case "REFERENCE":
                            node.metadata["reference_inputs"] += 1
                        case "REJECT":
                            node.metadata["reject_inputs"] += 1
                        case _:
                            raise ValueError(f"Unknown link type {link.type}")

            for child in node.metadata["children"]:
                if not isinstance(child, StageNode):
                    continue

                links = self.adj[node][child]
                for link in links:
                    match link.type:
                        case "PRIMARY":
                            node.metadata["primary_outputs"] += 1
                        case "REFERENCE":
                            node.metadata["reference_outputs"] += 1
                        case "REJECT":
                            node.metadata["reject_outputs"] += 1
                        case _:
                            raise ValueError(f"Unknown link type {link.type}")

        # Assign each node to a layer based on topological order
        for node in top_order:
            if not node.metadata["parents"]:
                # Source nodes reside in layer 0
                node.metadata["layer"] = 0
            else:
                # Non-source nodes reside in layer 1 + max(parents' layers)
                node.metadata["layer"] = max(parent.metadata["layer"] for parent in node.metadata["parents"]) + 1

        self.is_metadata_computed = True

    def get_connected_subgraphs(self):
        assert self.adj, "DAG is empty"

        # Compute mapping from nodes to their parents
        parents: dict[Node, list[Node]] = {}

        for parent_node, children_dict in self.adj.items():
            parents.setdefault(parent_node, [])
            for child_node, links in children_dict.items():
                parents.setdefault(child_node, [])
                parents[child_node].append(parent_node)

        subgraphs: list[DAG] = []
        visited = set()

        def dfs(cur_node: Node, dag: DAG):
            if cur_node in visited:
                return

            visited.add(cur_node)
            dag.adj[cur_node] = self.adj[cur_node]

            # DFS downstream
            for n in self.adj[cur_node].keys():
                dfs(n, dag)
            # DFS upstream
            for n in parents.get(cur_node):
                dfs(n, dag)

        for node in self.nodes():
            subgraph = DAG()
            dfs(node, subgraph)
            if subgraph.adj:
                subgraphs.append(subgraph)

        return subgraphs

    def __compute_in_degrees(self, condition: Callable[[Node], bool] = lambda _: True):
        in_degrees = {node: 0 for node in self.nodes()}
        for link in self.links():
            in_degrees[link.dest] += 1 if condition(link.src) else 0
        return in_degrees

    def __compute_out_degrees(self, condition: Callable[[Node], bool] = lambda _: True):
        out_degrees = {node: 0 for node in self.nodes()}
        for link in self.links():
            out_degrees[link.src] += 1 if condition(link.dest) else 0
        return out_degrees

    def __str__(self):
        out = ""
        for link in self.links():
            out += f"{link.src} -> {link.dest}\n"
        return out

    def __print_cycle(self, adj):
        """nodes: a list of nodes
        adj: a dict as adjacency graph
        output: string.
        """
        nodes = adj.keys()
        nodes_not_lead_to_deadend = set()

        for node in nodes:
            stack = [(node, [])]

            while len(stack) != 0:
                curr_node, visited = stack.pop()

                if curr_node in nodes_not_lead_to_deadend:
                    # since this node will not lead to any deadend, we ignore this node
                    continue

                if curr_node in visited:
                    # if current node is already visited in current path, we detected a cycle.
                    path = visited + [curr_node]
                    cycle = path[path.index(path[-1]) :]
                    cycle = [n.var_name for n in cycle]
                    cycle[0] = "\033[31m" + cycle[0] + "\033[0m"
                    cycle[-1] = "\033[31m" + cycle[-1] + "\033[0m"
                    return "[" + " >> ".join(cycle) + "]"

                if len(adj[node]) == 0:
                    # if this node has no outdegree. we ignore
                    nodes_not_lead_to_deadend.add(node)
                    continue

                for adj_node in adj[curr_node]:
                    if adj_node not in nodes_not_lead_to_deadend:
                        stack.append((adj_node, visited + [curr_node]))

        # if we got this far, this means our algorithm didn't detect any cycle. Though this shouldn't be the case.
        return ""

class StageNode(Node):
    def __init__(
        self,
        dag: "DAG",
        configuration: pydantic.BaseModel,
        *,
        label: str = None,
    ):
        super().__init__(dag)
        self.label = label
        self.configuration = configuration

    def _get_node_type(self) -> str | None:
        return self.configuration.node_type

    def _get_node_label(self) -> str | None:
        return self.label

    def _get_op_name(self) -> str | None:
        return self.configuration.op_name

    def _get_image(self) -> str:
        return self.configuration.image

    def _get_node_params(self) -> dict[str, Any] | None:
        return self.configuration.get_parameters_props()

    def _get_advanced_params(self) -> dict[str, Any] | None:
        return self.configuration.get_advanced_props()

    def _get_input_port_params(self, link=None) -> dict[str, Any] | None:
        if link:
            return self.configuration.get_input_ports_props(link)
        return self.configuration.get_input_ports_props()

    def _get_output_port_params(self, link=None) -> dict[str, Any] | None:
        if link:
            return self.configuration.get_output_ports_props(link)
        return self.configuration.get_output_ports_props()

    def _get_source_connection_params(self) -> dict[str, Any] | None:
        return self.configuration.get_source_props()

    def _get_target_connection_params(self):
        return self.configuration.get_target_props()

    def _get_max_primary_inputs(self) -> int:
        in_card = self.configuration.get_input_cardinality()
        return in_card["max"]

    def _get_min_primary_inputs(self) -> int:
        in_card = self.configuration.get_input_cardinality()
        return in_card["min"]

    def _get_max_primary_outputs(self) -> int:
        out_card = self.configuration.get_output_cardinality()
        return out_card["max"]

    def _get_min_primary_outputs(self) -> int:
        out_card = self.configuration.get_output_cardinality()
        return out_card["min"]

    def _get_min_reject_outputs(self) -> int:
        try:
            return self.configuration.get_app_data_props()["datastage"]["minRejectOutputs"]
        except AttributeError | KeyError:
            return super()._get_min_reject_outputs()

    def _get_max_reject_outputs(self) -> int:
        try:
            return self.configuration.get_app_data_props()["datastage"]["maxRejectOutputs"]
        except AttributeError | KeyError:
            return super()._get_max_reject_outputs()

    def _get_min_reference_inputs(self) -> int:
        try:
            return self.configuration.get_app_data_props()["datastage"]["minReferenceInputs"]
        except AttributeError | KeyError:
            return super()._get_min_reference_inputs()

    def _get_max_reference_inputs(self) -> int:
        try:
            return self.configuration.get_app_data_props()["datastage"]["maxReferenceInputs"]
        except AttributeError | KeyError:
            return super()._get_max_reference_inputs()

    def _get_rcp(self) -> bool | None:
        return self.configuration.runtime_column_propagation

    def _get_acp(self) -> bool | None:
        # Added this check as not every model has the `auto_column_propagation` property, and accessing it causes an error
        if hasattr(self.configuration, "auto_column_propagation"):
            return self.configuration.auto_column_propagation

    def _get_connection_params(self, location) -> dict[str, Any] | None:
        if location == "both":
            return {
                **self.configuration.get_source_props(),
                **self.configuration.get_target_props(),
            }
        elif location == "source":
            return self.configuration.get_source_props()
        elif location == "target":
            return self.configuration.get_target_props()
        return None

    def _get_connection_name(self) -> str | None:
        return self.configuration.connection.name

    def _get_project_id(self) -> str | None:
        return self.configuration.connection.proj_id

    def _get_connection_id(self) -> str | None:
        return self.configuration.connection.asset_id

    def __str__(self):
        return f"<{type(self).__name__}>"



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing StreamSets Flow Models."""

import inflection
import json
import logging
import re
import requests
import textwrap
from collections import defaultdict, namedtuple
from copy import deepcopy
from functools import cached_property
from ibm_watsonx_data_integration.common.constants import SUPPORTED_FLOWS
from ibm_watsonx_data_integration.common.exceptions import (
    StageNotFoundError,
    UniqueStageNameNotFoundError,
)
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from ibm_watsonx_data_integration.cpd_models.flow_model import Flow, PayloadExtender
from ibm_watsonx_data_integration.services.streamsets.data import create_pipeline_definition_for_new_flow
from ibm_watsonx_data_integration.services.streamsets.exceptions import StageDefinitionNotFound
from ibm_watsonx_data_integration.services.streamsets.models import Engine, Environment
from ibm_watsonx_data_integration.services.streamsets.models.configuration import Configuration
from ibm_watsonx_data_integration.services.streamsets.models.engine_model import LibraryDefinitions
from pydantic import ConfigDict, Field, PrivateAttr
from typing import TYPE_CHECKING, Any, ClassVar
from typing_extensions import override
from urllib.parse import parse_qs, urlparse
from uuid import uuid4

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.cpd_models import Project
    from ibm_watsonx_data_integration.platform import Platform

logger = logging.getLogger()

StageClassDefault = namedtuple("StageClassDefaults", ["class_name", "attributes"])
StageData = namedtuple("StageData", ["definition", "instance"])
StageConfigurationProperty = namedtuple("StageConfigurationProperty", ["config_name"])
StageServiceProperty = namedtuple("StageServiceProperty", ["service_name", "config_name"])


class Stage(BaseModel):
    """Class for a stage."""

    instance_name: str = Field(alias="instanceName", repr=False, frozen=True)
    library: str = Field(alias="library", repr=False, frozen=True)
    stage_name: str = Field(alias="stageName", repr=False, frozen=True)
    stage_version: str = Field(alias="stageVersion", repr=False, frozen=True)
    ui_info: dict = Field(alias="uiInfo", repr=False, frozen=True)
    input_lanes: list[str] = Field(alias="inputLanes", repr=False, default=list())
    output_lanes: list[str] = Field(alias="outputLanes", repr=False, default=list())
    event_lanes: list[str] = Field(alias="eventLanes", repr=False, default=list())
    services_data: list[dict] = Field(alias="services", repr=False)
    configuration_data: list[dict] = Field(alias="configuration", repr=False)

    def __init__(
        self,
        pipeline_definition: "PipelineDefinition",
        output_streams: int = 0,
        supported_connection_types: list[str] | None = None,
        attributes: dict | None = None,
        **stage_json: any,
    ) -> None:
        """The __init__ for the class.

        Args:
            pipeline_definition: Pipeline Definition of the flow this stage belongs to
            output_streams: The number of output streams for the stage, if 0 then it is inferred from output lanes
            supported_connection_types: The connection types supported by the stage.
            attributes: Any attributes specific to the stage.
            stage_json: All other properties used by BaseModel
        """
        super().__init__(**stage_json)
        self._attributes = attributes or {}
        self._pipeline_definition = pipeline_definition

        self._supported_connection_types = supported_connection_types
        self._output_streams = output_streams
        # for adding outputs to the stage
        self._output_lane_idx = 0
        # output lanes need to be set sometimes
        if self.output_lanes:
            self._output_streams = len(self.output_lanes)
        else:
            for i in range(self._output_streams):
                self.output_lanes.append(f"{self.instance_name}OutputLane{str(uuid4())}".replace("-", "_"))
        # create the configuration and service configurations
        configurations = [self.configuration_data]
        self._services = dict()
        for service in self.services_data:
            self._services[service["service"]] = Configuration(service["configuration"])
            configurations.append(service["configuration"])
        self._configurations = Configuration(configurations, id_to_remap=self._create_id_to_remap())

        # Add a docstring to show stage attributes when help() is run on a stage instance.
        # Attributes will be printed in two columns.
        attrs_ = list(sorted(self._attributes.keys()))
        split = int(len(attrs_) / 2)
        self.__class__.__doc__ = textwrap.dedent(
            """
            {class_name}: {label}
            Attributes:
            {attributes}
            """
        ).format(
            class_name=self.__class__.__name__,
            label=self.stage_name,
            attributes="\n".join(f"{first:<60}{second:<60}" for first, second in zip(attrs_[:split], attrs_[split:])),
        )

    def _create_id_to_remap(self) -> dict:
        """Constructs a mapping of stage_configuration_backend_name: stage_configuration_human_readable_name."""
        id_to_remap = {}
        for name, value in self._attributes.items():
            config_properties = [value] if not isinstance(value, list) else value
            for config_property in config_properties:
                id_to_remap[name] = config_property.config_name
        return id_to_remap

    @property
    def configuration(self) -> Configuration:
        """The stage's configurations."""
        return self._configurations

    def _add_output(self, *stages: "Stage", is_event: bool = False, predicate: dict = None) -> None:
        """Adds stages to the output of this stage."""
        if not stages:
            raise ValueError("Stages cannot be empty.")
        if predicate and not isinstance(self, StageWithPredicates):
            raise TypeError("Cannot pass predicate argument for a stage without predicates.")
        if is_event and predicate:
            raise ValueError("Choose one of is_event or predicate.")
        if (
            predicate
            and not predicate.get("outputLane")
            and not next(output_lane for output_lane in self.output_lanes if output_lane == predicate.get("outputLane"))
        ):
            raise TypeError("Pass a valid predicate with an existing output lane.")
        if not predicate and not is_event and self._output_lane_idx >= self._output_streams:
            raise ValueError("Stage does not produce any outputs.")

        if is_event:
            if not self.event_lanes:
                lane = f"{self.instance_name}_EventLane"
                self.event_lanes.append(lane)
            else:
                lane = self.event_lanes[0]
        elif predicate:
            lane = predicate.get("outputLane")
        else:
            lane = self.output_lanes[self._output_lane_idx]
            if self._output_lane_idx < self._output_streams - 1:
                self._output_lane_idx += 1
        for stage in stages:
            stage.input_lanes.append(lane)

    def connect_output_to(self, *stages: "Stage", predicate: dict = None) -> None:
        """Connect other stages to the output of this stage.

        Args:
            stages: Stages to connect to this stage
            predicate: Whether to use any specific predicate when connecting (only applicable to some stages)
        """
        self._add_output(*stages, is_event=False, predicate=predicate)

    def connect_input_to(self, *stages: "Stage", predicate: dict = None) -> None:
        """Connect other stages to the input of this stage.

        Args:
            stages: Stages to connect to this stage
            predicate: Whether to use any specific predicate when connecting
        """
        for stage in stages:
            stage.connect_output_to(self, predicate=predicate)

    def connect_event_to(self, *stages: "Stage") -> None:
        """Connect other stages to the event output of this stage.

        Args:
            stages: Stages to connect to this stage
        """
        self._add_output(*stages, is_event=True)

    def _remove_input(self, *stages: "Stage", is_event: bool = False) -> None:
        """Disconnect the inputs from this stage."""
        lane_identifier = "_EventLane" if is_event else "OutputLane"
        instance_names = [stage.instance_name for stage in stages]
        self.input_lanes = list(
            filter(lambda lane: lane.split(lane_identifier)[0] not in instance_names, self.input_lanes)
        )

    def disconnect_input_from(self, *stages: "Stage", is_event: bool = False) -> None:
        """Disconnect the input of this stage from the output of the provided stages.

        Args:
            stages: Stages to disconnect from this stage.
            is_event: Whether we are disconnecting events.
        """
        self._remove_input(*stages, is_event=is_event)

    def disconnect_output_from(self, *stages: "Stage") -> None:
        """Disconnect the output of this stage from the provided stages.

        Args:
            stages: Stages to disconnect from this stage.
        """
        for stage in stages:
            stage.disconnect_input_from(self)

    def disconnect_event_from(self, *stages: "Stage") -> None:
        """Disconnect the event output of this stage from the provided stages.

        Args:
            stages: Stages to disconnect from this stage.
        """
        for stage in stages:
            stage.disconnect_input_from(self, is_event=True)

    @property
    def outputs(self) -> list:
        """All the stages connected to this stages output."""
        return list(
            filter(
                lambda stage: any([lane in self.output_lanes for lane in stage.input_lanes]),
                self._pipeline_definition.stages,
            )
        )

    @property
    def inputs(self) -> list:
        """All the stages connected to this stages input."""
        return list(
            filter(
                lambda stage: any(
                    [(lane in stage.output_lanes) or (lane in stage.event_lanes) for lane in self.input_lanes]
                ),
                self._pipeline_definition.stages,
            )
        )

    @property
    def events(self) -> list:
        """All the stages connected to this stages event output."""
        return list(
            filter(
                lambda stage: any([lane in self.event_lanes for lane in stage.input_lanes]),
                self._pipeline_definition.stages,
            )
        )

    @property
    def type(self) -> str:
        """The type of this stage."""
        return self.ui_info.get("stageType")

    def __getattr__(self, name: str) -> any:
        """Get an attribute of this stage, includes any special attributes for this stage."""
        if name != "_attributes" and name in self._attributes:
            attr_value = self._attributes.get(name)
            config_properties = [attr_value] if not isinstance(attr_value, list) else attr_value

            for config_property in config_properties:
                if (
                    isinstance(config_property, StageConfigurationProperty)
                    and config_property.config_name in self.configuration
                ):
                    return self.configuration[config_property.config_name]
                elif (
                    isinstance(config_property, StageServiceProperty) and config_property.service_name in self._services
                ):
                    return self._services[config_property.service_name][config_property.config_name]

        return super().__getattr__(name)

    def __setattr__(self, name: str, value: any) -> None:
        """Set an attribute of this stage, includes any special attributes for this stage."""
        if name != "_attributes" and name in self._attributes:
            attr_value = self._attributes.get(name)
            config_properties = [attr_value] if not isinstance(attr_value, list) else attr_value

            found_config = False
            for config_property in config_properties:
                if (
                    isinstance(config_property, StageConfigurationProperty)
                    and config_property.config_name in self.configuration
                ):
                    found_config = True
                    self.configuration[config_property.config_name] = value
                elif (
                    isinstance(config_property, StageServiceProperty) and config_property.service_name in self._services
                ):
                    found_config = True
                    self._services[config_property.service_name][config_property.config_name] = value

            if not found_config:
                raise AttributeError(f"'{type(self).__name__!r}' has no attribute '{name!r}'.")
        else:
            super().__setattr__(name, value)

    def __dir__(self) -> list[str]:
        """The dir of a stage."""
        return sorted(list(super().__dir__()) + list(self._attributes.keys()))

    def __str__(self) -> str:
        """This stage's representation as a string."""
        from_pydantic = super().__str__()
        return self.instance_name + from_pydantic[len(self.__class__.__name__) :]

    def __repr__(self) -> str:
        """The stage's representation."""
        return str(self)


class StageWithPredicates(Stage):
    """Class for a stage with predicates."""

    def __init__(self, variable_output_drive: str, *args: any, **kwargs: any) -> None:
        """The __init__ for stages with predicates.

        Args:
            variable_output_drive: The configuration key for where the predicates are located.
            args: Arguments for initializing a Stage.
            kwargs: Keyword arguments for initializing a Stage.
        """
        super().__init__(*args, **kwargs)
        self._variable_output_drive = variable_output_drive
        if not self.predicates:
            self.predicates = ["default"]

    def _prepare_predicates(self, predicates: list[dict | str]) -> list[dict]:
        """Create a set of predicates from a given set of conditions."""
        formatted_predicates = []
        for predicate in predicates:
            if isinstance(predicate, str):
                predicate = {"predicate": predicate}
            output_lane = predicate.get("outputLane", f"{self.instance_name}OutputLane{str(uuid4())}".replace("-", "_"))
            if not predicate.get("predicate", None):
                raise ValueError("Output Lane drives should have a predicate key.")
            if not predicate.get("outputLane", None):
                output_lane_config = {"outputLane": output_lane}
                # We make sure that the predicate is in the form {'outputLane': 'value', 'predicate': 'value'} to be
                # consistent with the UI
                predicate = {**output_lane_config, **predicate}
            formatted_predicates.append(predicate)
        return formatted_predicates

    def _disconnect_and_remove_predicate(self, i: int) -> None:
        """Disconnect and remove a predicate at location i."""
        deleted_predicate = self.predicates.pop(i)
        output_lane = deleted_predicate["outputLane"]

        output_stages_to_disconnect = [stage for stage in self.outputs if output_lane in stage.input_lanes]
        self.disconnect_output_from(*output_stages_to_disconnect)

        self.output_lanes.remove(output_lane)
        self._output_streams -= 1

    @property
    def predicates(self) -> list[dict]:
        """Get the predicate list for this stage."""
        return self.configuration[self._variable_output_drive]

    @predicates.setter
    def predicates(self, predicates: list[dict | str]) -> None:
        """Set the predicates for this stage."""
        predicates = self._prepare_predicates(predicates)
        # Create default condition if not present
        if not [predicate for predicate in predicates if predicate["predicate"] == "default"]:
            default_predicate = self._prepare_predicates(["predicate"])
            predicates.extend(default_predicate)
        if not predicates[-1]["predicate"] == "default":
            raise ValueError("The default predicate must be placed at the end of the list.")
        self.disconnect_output_from(*self.outputs)  # disconnect all output stages
        self._output_streams = 0
        self.output_lanes = []
        self._output_lane_idx = 0
        # Create output lanes in parent stage for each predicate
        for predicate in predicates:
            self.output_lanes.append(predicate["outputLane"])
            self._output_streams += 1
        self.configuration[self._variable_output_drive] = predicates

    def add_predicates(self, predicates: list[dict | str]) -> None:
        """Add a predicate.

        Example:
            stage.add_predicates(['>0'])
            stage.add_predicates([{'predicate':'>0', 'outputLane':'lane1'}])
            stage.add_predicates(['>0' ,'=0'])

        Args:
            predicates: The list of predicates to add.

        Raises:
            ValueError: If predicates is not a list.
        """
        if not isinstance(predicates, list):
            raise ValueError("Predicates should be a list.")
        formatted_predicates = self._prepare_predicates(predicates)
        for new_predicate in formatted_predicates:
            self.output_lanes.insert(0, new_predicate["outputLane"])
            self.predicates.insert(0, new_predicate)
            self._output_streams += 1

    def remove_predicate(self, predicate: dict) -> None:
        """Remove a predicate.

        Example:
            stage.remove_predicate(stage.predicates[0])
            stage.remove_predicate({'predicate':'>0', 'outputLane':'lane1'})

        Args:
            predicate: The predicate to delete as a dictionary including the outputLane.

        Raises:
            ValueError: If predicates is not specified or can't find its target.
        """
        clean_predicate = next(iter(self._prepare_predicates([predicate])), {})
        if not clean_predicate:
            raise ValueError("Need to specify a predicate")
        if clean_predicate.get("predicate") == "default":
            raise ValueError("Can't delete the default predicate.")
        for i, found_predicate in enumerate(self.predicates):
            if found_predicate == clean_predicate:
                self._disconnect_and_remove_predicate(i)
                return
        raise ValueError("Can't find target predicate in the predicates list.")


class PipelineDefinition(BaseModel):
    """Pipeline Definition of a flow in an engine."""

    _NODE_TYPES: ClassVar[dict] = {
        "origin": "SOURCE",
        "destination": "TARGET",
        "executor": "EXECUTOR",
        "processor": "PROCESSOR",
    }

    @staticmethod
    def get_attribute(config_definition: dict) -> tuple[str, str]:
        """Gets the attribute name for a configuration using its definition in a human-readable format."""
        config_name = config_definition.get("name")
        config_label = config_definition.get("label")
        if config_label:
            replacements = [(r"[\s-]+", "_"), (r"&", "and"), (r"/sec", "_per_sec"), (r"_\((.+)\)", r"_in_\1")]
            attribute_name = config_label.lower()
            for pattern, replacement in replacements:
                attribute_name = re.sub(pattern, replacement, attribute_name)
        else:
            attribute_name = inflection.underscore(config_definition["fieldName"])
        return attribute_name, config_name

    @staticmethod
    def get_color_icon_from_stage_definition(stage_definition: dict, stage_types: dict) -> str:
        """Create `colorIcon` value from stage definition.

        Args:
            stage_definition: stage definition from pipeline definitions
            stage_types: The PipelineDefinition._NODE_TYPES
        """
        reversed_stage_types = {v: k for k, v in stage_types.items()}
        if stage_definition.get("type") not in reversed_stage_types or not stage_definition.get("label"):
            return ""  # an empty string shows a greyed out image on the UI
        stage_processor_type = reversed_stage_types[stage_definition["type"]].capitalize()
        # space or slash should be delimiters
        escaped_stage_labels = re.split(r"[ /]", stage_definition["label"])
        color_icon_name = "_".join([stage_processor_type] + escaped_stage_labels) + ".png"
        return color_icon_name

    @staticmethod
    def stage_configuration_name(stage_json: dict) -> str:
        """Gets a stages configuration name, the name it has in a flow's configuration."""
        return f"{stage_json['library']}::{stage_json['stageName']}::{stage_json['stageVersion']}"

    schema_version: int = Field(alias="schemaVersion", repr=False)
    version: int = Field(alias="version", repr=False)
    pipeline_id: str = Field(alias="pipelineId", repr=False)
    title: str = Field(alias="title", repr=False)
    description: str = Field(alias="description", repr=False)
    uuid: str = Field(alias="uuid", repr=False)
    ui_info: dict = Field(alias="uiInfo", repr=False)
    fragments: list = Field(alias="fragments", repr=False)
    info: dict = Field(alias="info", repr=False)
    metadata: dict | None = Field(alias="metadata", repr=False, default=None)
    valid: bool = Field(alias="valid", repr=False)
    issues: dict = Field(alias="issues", repr=False)
    previewable: bool = Field(alias="previewable", repr=False)

    configuration_data: list[dict] = Field(alias="configuration", repr=False)

    stages_data: list[dict] = Field(alias="stages", repr=False)
    error_stage_data: dict | None = Field(alias="errorStage", repr=False)
    stats_aggregator_stage_data: dict | None = Field(alias="statsAggregatorStage", repr=False)
    start_event_stages_data: list[dict] | None = Field(default=None, alias="startEventStages", repr=False)
    stop_event_stages_data: list[dict] | None = Field(default=None, alias="stopEventStages", repr=False)
    test_origin_stages_data: dict | None = Field(default=None, alias="testOriginStage", repr=False)

    def __init__(self, flow: "StreamsetsFlow", **pipeline_definition_json: any) -> None:
        """The __init__ for the pipeline definition.

        Args:
            flow: The parent flow object
            pipeline_definition_json: Pipeline definition from the engine.
        """
        super().__init__(**pipeline_definition_json)
        self._flow = flow

        # lazy load library definitions
        self.__library_definitions: LibraryDefinitions | None = None

        self.__all_stages = None

        # for all the stage objects in the flow currently.
        self.__current_stages = None

        # all the special stages
        self.__error_stage = None
        self.__stats_aggregator_stage = None
        self.__start_event_stage = None
        self.__stop_event_stage = None
        self.__test_origin_stage = None

    @property
    def _library_definitions(self) -> LibraryDefinitions:
        if self.__library_definitions is None:
            if self._flow.engine is not None:
                # if possible try to get the library definitions from the engine
                try:
                    self.__library_definitions = self._flow.engine.library_definitions
                except requests.exceptions.HTTPError:
                    pass
            if self.__library_definitions is None:
                # if the library definitions is still None, we either don't have a running engine
                # or we faced an error trying to get the definitions from the engine
                # in this case, we should use a flow's engine_version
                self.__library_definitions = LibraryDefinitions.for_engine_version(
                    engine_version=self._flow.engine_version, platform=self._flow._project._platform
                )

        return self.__library_definitions

    @property
    def _all_stages(self) -> dict:
        if self.__all_stages is None:
            self.__all_stages = {}

            for stage_definition in self._library_definitions.stages:
                stage_name = stage_definition["name"]

                # Get configuration definition attributes
                attributes = defaultdict(list)
                for config_definition in stage_definition["configDefinitions"]:
                    attribute_name, config_name = self.get_attribute(config_definition)
                    attributes[attribute_name].append(StageConfigurationProperty(config_name=config_name))

                for service in stage_definition.get("services"):
                    service_name = service["service"]
                    service_definition = next(
                        (
                            service_def
                            for service_def in self._library_definitions.services
                            if service_def["provides"] == service_name
                        ),
                        None,
                    )

                    if service_definition is None:
                        raise StageDefinitionNotFound(f"Could not find definition of service {service_name}")

                    for config_definition in service_definition["configDefinitions"]:
                        attribute_name, config_name = self.get_attribute(config_definition)
                        attributes[attribute_name].append(
                            StageServiceProperty(service_name=service_name, config_name=config_name)
                        )
                # Invoke the StageWithPredicates class IFF the stage's output streams are governed by lane predicates
                variable_output_drive = stage_definition.get("outputStreamsDrivenByConfig", None)
                if variable_output_drive and variable_output_drive == "lanePredicates":
                    self.__all_stages[stage_name] = StageClassDefault(
                        class_name=StageWithPredicates, attributes=attributes
                    )
                else:
                    self.__all_stages[stage_name] = StageClassDefault(class_name=Stage, attributes=attributes)

        return self.__all_stages

    @property
    def configuration(self) -> Configuration:
        """The configuration for the pipeline."""
        config_definitions = self._library_definitions.pipeline[0]["configDefinitions"]

        mapping = {}
        for config_definition in config_definitions:
            attribute_name, config_name = self.get_attribute(config_definition)
            mapping[attribute_name] = config_name

        return Configuration(
            configuration=self.configuration_data,
            id_to_remap=mapping,
        )

    def _get_stage_data(
        self, label: str | None = None, name: str | None = None, type: str | None = None, library: str | None = None
    ) -> list["StageData"]:
        """Get a stages definition and new instance based on parameters provided.

        Raises:
            ValueError: If both label and name are provided, or if neither is provided.
            StageNotFoundError: If a stage wasn't found.
        """
        if label and name:
            raise ValueError("Use `label` or `name`, not both.")
        elif not label and not name:
            raise ValueError("Either `label` or `name` must be specified.")
        stages = [
            stage
            for stage in self._library_definitions.stages
            if (
                (label and stage["label"] == label or name and stage["name"] == name)
                and (not library or stage["library"] == library)
                and (not type or stage["type"] == self._NODE_TYPES.get(type, type))
            )
        ]
        if not stages:
            raise StageNotFoundError(f"Could not find a stage ({label or name}).")
        return [StageData(definition=stage, instance=self._get_new_stage_instance(stage)) for stage in stages]

    def _get_new_stage_instance(self, stage: dict) -> dict:
        """Create a new instance of a stage's json based on definition."""
        ui_info = dict(
            colorIcon=self.get_color_icon_from_stage_definition(stage, self._NODE_TYPES),
            description="",
            label=self._get_stage_label(stage),
            xPos=0,
            yPos=0,
            stageType=stage.get("type"),
        )
        stage_instance = dict(
            instanceName=self._get_stage_instance_name(stage),
            library=stage.get("library"),
            stageName=stage.get("name"),
            stageVersion=stage.get("version"),
            configuration=list(),
            uiInfo=ui_info,
            inputLanes=list(),
            outputLanes=list(),
            eventLanes=list(),
        )
        stage_instance["configuration"] = [
            self._set_default_value_for_config(config_definition, stage_instance)
            for config_definition in stage["configDefinitions"]
        ]
        stage_instance["services"] = [
            {
                "service": service_definition["provides"],
                "serviceVersion": service_definition["version"],
                "configuration": [
                    self._set_default_value_for_config(config_definition, None)
                    for config_definition in service_definition["configDefinitions"]
                ],
            }
            for service in stage["services"]
            for service_definition in self._library_definitions.services
            if service_definition["provides"] == service["service"]
        ]
        # Propagate RUNTIME configuration injected by the stage.
        for stage_instance_service in stage_instance["services"]:
            for service in stage["services"]:
                if stage_instance_service["service"] == service["service"]:
                    Configuration(stage_instance_service["configuration"]).update(service["configuration"])
        return stage_instance

    def _get_stage_instance_name(self, stage: dict) -> str:
        """Get the instance name for a new stage. Since new instance names increment by value of 1."""
        stage_name = re.sub("[^0-9A-Za-z_]", "", stage["label"])
        if stage["errorStage"]:
            return f"{stage_name}_ErrorStage"
        elif stage["statsAggregatorStage"]:
            return f"{stage_name}_StatsAggregatorStage"
        else:
            similar_instances = set(stage.instance_name for stage in self.stages)
            for i in range(1, 1001):  # upto 1000 stages in a sdc
                new_instance_name = f"{stage_name}_{i:0>2}"
                if new_instance_name not in similar_instances:
                    return new_instance_name
        raise UniqueStageNameNotFoundError("Couldn't find unique instance name for stage.")

    def _get_stage_label(self, stage: dict) -> str:
        """Gets the label of a stage."""
        if stage["errorStage"]:
            return "Error Records - {}".format(stage["label"])
        elif stage["statsAggregatorStage"]:
            return "Stats Aggregator - {}".format(stage["label"])
        else:
            similar_instances = sum(
                stage.get("label") in existing_stage.ui_info.get("label") for existing_stage in self.stages
            )
            return "{} {}".format(stage["label"], similar_instances + 1)

    def _set_default_value_for_config(self, config_definition: dict, stage_instance: dict | None) -> dict:
        # A port of pipelineService.js's setDefaultValueForConfig method (https://git.io/vSh3W).
        config = {"name": config_definition["name"], "value": config_definition["defaultValue"]}
        if config_definition["type"] == "MODEL":
            if config_definition["model"]["modelType"] == "FIELD_SELECTOR_MULTI_VALUE" and not config["value"]:
                config["value"] = []
            # We don't follow the logic for PREDICATE as that assumes that the stage already have output lanes.
            # However, this is called at the time of stage initialization on our side and the stage output lanes
            # does not exist until user explicitly connects stages together.
            elif config_definition["model"]["modelType"] == "LIST_BEAN" and config["value"] is None:
                config["value"] = [
                    {
                        self._set_default_value_for_config(model_config_definition, stage_instance)[
                            "name"
                        ]: self._set_default_value_for_config(model_config_definition, stage_instance).get("value")
                        for model_config_definition in config_definition["model"]["configDefinitions"]
                        if (
                            self._set_default_value_for_config(model_config_definition, stage_instance).get("value")
                            is not None
                        )
                    }
                ]
        elif config_definition["type"] == "BOOLEAN" and config["value"] is None:
            config["value"] = False
        elif config_definition["type"] == "LIST" and not config["value"]:
            config["value"] = []
        elif config_definition["type"] == "MAP" and not config["value"]:
            config["value"] = []
        return config

    def add_stage(
        self, label: str | None = None, name: str | None = None, type: str | None = None, library: str | None = None
    ) -> Stage | StageWithPredicates:
        """Add a new stage to the pipeline.

        Args:
            label: Label of the stage
            name: Name of the stage
            type: Type of stage
            library: Library the stage belongs to
        """
        # get the data for the stage
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, type=type, library=library)
            if stage_data.definition.get("errorStage") is False
        )
        # get output streams, predicates and supported connection types
        output_streams = stage_definition.get("outputStreams", 0)
        variable_output_drive = stage_definition.get("outputStreamsDrivenByConfig", None)
        extra_args = {"variable_output_drive": variable_output_drive} if variable_output_drive else {}
        supported_connection_types = [
            config["connectionType"]
            for config in stage_definition["configDefinitions"]
            # we want all the fields that have connectionType
            if config.get("connectionType")
        ]
        # get the appropriate class with attributes and initialize it.
        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        stage_object = stage_class(
            pipeline_definition=self,
            output_streams=output_streams,
            supported_connection_types=supported_connection_types,
            attributes=stage_attributes,
            **stage_instance,  # stage_json
            **extra_args,
        )
        self.stages.append(stage_object)
        return stage_object

    def remove_stage(self, stage: Stage) -> None:
        """Removed a stage from the flow.

        Args:
            stage: Stage to remove.
        """
        # remove it from the underlying json
        self.stages = [
            current_stage for current_stage in self.stages if current_stage.instance_name != stage.instance_name
        ]
        # disconnect everything
        stage.disconnect_input_from(*self.stages)
        stage.disconnect_output_from(*self.stages)
        stage.disconnect_event_from(*self.stages)

    def _convert_stage_from_json_to_obj(self, stage_json: dict) -> Stage:
        """Take a stage dict from pipeline definition and convert to a stage class."""
        stage_definition = next(
            (
                stage_def
                for stage_def in self._library_definitions.stages
                if stage_def["name"] == stage_json["stageName"]
            ),
            {},
        )

        if not stage_definition:
            logger.debug("Definitions not found for stage", extra=dict(stage_name=stage_json["stageName"]))

        supported_connection_types = [
            config["connectionType"]
            for config in stage_definition.get("configDefinitions", [])
            if config.get("connectionType")
        ]
        extra_args = {}
        variable_output_drive = stage_definition.get("outputStreamsDrivenByConfig")
        if variable_output_drive == "lanePredicates":
            extra_args.update({"variable_output_drive": variable_output_drive})

        stage_class, stage_attributes = self._all_stages.get(
            stage_json["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        stage_object = stage_class(
            pipeline_definition=self,
            supported_connection_types=supported_connection_types,
            attributes=stage_attributes,
            **stage_json,
            **extra_args,
        )

        return stage_object

    @property
    def stages(self) -> list[Stage | StageWithPredicates]:
        """The stages belonging to the flow."""
        if self.__current_stages is None:
            self.__current_stages = [
                self._convert_stage_from_json_to_obj(stage_json=stage_json) for stage_json in self.stages_data
            ]
        return self.__current_stages

    @stages.setter
    def stages(self, val: list[Stage]) -> None:
        self.__current_stages = val

    @property
    def error_stage(self) -> Stage | None:
        """The error stage of the flow."""
        if self.__error_stage is None:
            if not self.error_stage_data:
                return None
            else:
                self.__error_stage = self._convert_stage_from_json_to_obj(self.error_stage_data)
        return self.__error_stage

    @property
    def stats_aggregator_stage(self) -> Stage | None:
        """The stats aggregator stage of the flow."""
        if self.__stats_aggregator_stage is None:
            if not self.stats_aggregator_stage_data:
                return None
            else:
                self.__stats_aggregator_stage = self._convert_stage_from_json_to_obj(self.stats_aggregator_stage_data)
        return self.__stats_aggregator_stage

    @property
    def start_event(self) -> Stage | None:
        """The start event stage of the flow."""
        if self.__start_event_stage is None:
            if not self.start_event_stages_data or not self.start_event_stages_data[0]:
                return None
            else:
                self.__start_event_stage = self._convert_stage_from_json_to_obj(self.start_event_stages_data[0])
        return self.__start_event_stage

    @property
    def stop_event(self) -> Stage | None:
        """The stop event stage of the flow."""
        if self.__stop_event_stage is None:
            if not self.stop_event_stages_data or not self.stop_event_stages_data[0]:
                return None
            else:
                self.__stop_event_stage = self._convert_stage_from_json_to_obj(self.stop_event_stages_data[0])
        return self.__stop_event_stage

    @property
    def test_origin(self) -> Stage | None:
        """The test origin stage of the flow."""
        if self.__test_origin_stage is None:
            if not self.test_origin_stages_data:
                return None
            else:
                self.__test_origin_stage = self._convert_stage_from_json_to_obj(self.test_origin_stages_data)
        return self.__test_origin_stage

    def set_error_stage(self, label: str | None = None, name: str | None = None, library: str | None = None) -> Stage:
        """Set the error stage of the flow.

        Args:
            label: Label of the stage.
            name: Name of the stage.
            library: Library the stage belongs to.
        """
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, library=library)
            if stage_data.definition.get("errorStage") is True
        )

        self.error_stage_data = stage_instance
        self.configuration["badRecordsHandling"] = self.stage_configuration_name(stage_instance)
        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        self.__error_stage = stage_class(
            pipeline_definition=self,
            output_streams=stage_definition.get("outputStreams", 0),
            **stage_instance,
        )
        return self.__error_stage

    def set_stats_aggregator_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the stats aggregator stage of the flow.

        Args:
            label: Label of the stage.
            name: Name of the stage.
            library: Library the stage belongs to.
        """
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, library=library)
            if stage_data.definition.get("statsAggregatorStage") is True
        )

        self.stats_aggregator_stage_data = stage_instance
        self.configuration["statsAggregatorStage"] = self.stage_configuration_name(stage_instance)
        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        self.__stats_aggregator_stage = stage_class(
            pipeline_definition=self, output_streams=stage_definition.get("outputStreams", 0), **stage_instance
        )
        return self.__stats_aggregator_stage

    def set_start_event_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the start event stage of the flow.

        Args:
            label: Label of the stage.
            name: Name of the stage.
            library: Library the stage belongs to.
        """
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, library=library)
            if stage_data.definition.get("pipelineLifecycleStage") is True
        )

        # We need instanceName in the form <instance_name>_StartEventStage and not <instance_name>_01
        instance_name_split_list = stage_instance["instanceName"].split("_")
        instance_name_split_list[-1] = "StartEventStage"
        stage_instance["instanceName"] = "_".join(instance_name_split_list)
        self.start_event_stages_data = [stage_instance]
        self.configuration["startEventStage"] = self.stage_configuration_name(stage_instance)

        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        self.__start_event_stage = stage_class(
            pipeline_definition=self, output_streams=stage_definition.get("outputStreams", 0), **stage_instance
        )
        return self.__start_event_stage

    def set_stop_event_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the stop event stage of the flow.

        Args:
            label: Label of the stage.
            name: Name of the stage.
            library: Library the stage belongs to.
        """
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, library=library)
            if stage_data.definition.get("pipelineLifecycleStage") is True
        )

        # We need instanceName in the form <instance_name>_StopEventStage and not <instance_name>_01
        instance_name_split_list = stage_instance["instanceName"].split("_")
        instance_name_split_list[-1] = "StopEventStage"
        stage_instance["instanceName"] = "_".join(instance_name_split_list)
        self.stop_event_stages_data = [stage_instance]
        self.configuration["stopEventStage"] = self.stage_configuration_name(stage_instance)
        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        self.__stop_event_stage = stage_class(
            pipeline_definition=self, output_streams=stage_definition.get("outputStreams", 0), **stage_instance
        )
        return self.__stop_event_stage

    def set_test_origin_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the test origin stage of the flow.

        Args:
            label: Label of the stage.
            name: Name of the stage.
            library: Library the stage belongs to.
        """
        stage_instance, stage_definition = next(
            (stage_data.instance, stage_data.definition)
            for stage_data in self._get_stage_data(label=label, name=name, library=library)
            if stage_data.definition.get("errorStage") is True
        )

        # We need instanceName in the form <instance_name>_TestOriginStage and not <instance_name>_01
        instance_name_split_list = stage_instance["instanceName"].split("_")
        instance_name_split_list[-1] = "TestOriginStage"
        stage_instance["instanceName"] = "_".join(instance_name_split_list)
        self.test_origin_stages_data = stage_instance
        self.configuration["testOriginStage"] = self.stage_configuration_name(stage_instance)
        stage_class, stage_attributes = self._all_stages.get(
            stage_instance["stageName"], StageClassDefault(class_name=Stage, attributes={})
        )

        self.__test_origin_stage = stage_class(
            pipeline_definition=self, output_streams=stage_definition.get("outputStreams", 0), **stage_instance
        )
        return self.__test_origin_stage

    def _sort_stage_instances(self) -> list[Stage]:
        """A port of pipelineService.js's sortStageInstances (https://git.io/JmAoO)."""
        sorted_stages = list()
        removed_map = dict()
        produced_outputs = list()
        check = True
        iteration = 0

        while check:
            prior = len(sorted_stages)
            for stage in self.stages:
                if stage.instance_name not in removed_map:
                    already_produced = [item for item in produced_outputs if item in stage.input_lanes]
                    if len(already_produced) == len(stage.input_lanes):
                        produced_outputs += stage.output_lanes
                        produced_outputs += stage.event_lanes
                        removed_map[stage.instance_name] = True
                        sorted_stages.append(stage)
            iteration += 1
            if prior == len(sorted_stages) and iteration >= len(sorted_stages):
                check = False
                for stage in self.stages:
                    if stage.instance_name not in removed_map:
                        sorted_stages.append(stage)

        return sorted_stages

    def _auto_arrange_stages(self) -> None:
        """https://sor.bz/N3QcJ. Sets the x-pos and y-pos values of a stage so that they look organized on the UI."""
        sorted_stages = self._sort_stage_instances()

        x_pos = 60
        y_pos = 50
        lane_y_pos = {}
        lane_x_pos = {}

        for stage in sorted_stages:
            y = lane_y_pos.get(stage.input_lanes[0], 0) if len(stage.input_lanes) else y_pos
            x = lane_x_pos.get(stage.input_lanes[0], 0) + 220 if len(stage.input_lanes) else x_pos

            if len(stage.input_lanes) > 1:
                m_x = 0
                for input_lane in stage.input_lanes:
                    m_x = max(lane_x_pos.get(input_lane, 0), m_x)
                x = m_x + 220

            if stage.input_lanes and lane_y_pos.get(stage.input_lanes[0], 0):
                lane_y_pos[stage.input_lanes[0]] += 150

            if not y:
                y = y_pos

            if len(stage.output_lanes) > 1:
                for i, output_lane in enumerate(stage.output_lanes):
                    lane_y_pos[output_lane] = y - 10 + (130 * i)
                    lane_x_pos[output_lane] = x

                if y == y_pos:
                    y += 30 * len(stage.output_lanes)
            else:
                if len(stage.output_lanes):
                    lane_y_pos[stage.output_lanes[0]] = y
                    lane_x_pos[stage.output_lanes[0]] = x

                if len(stage.input_lanes) > 1 and y == y_pos:
                    y += 130

            if len(stage.event_lanes):
                lane_y_pos[stage.event_lanes[0]] = y + 150
                lane_x_pos[stage.event_lanes[0]] = x

            stage.ui_info["xPos"] = x
            stage.ui_info["yPos"] = y

            x_pos = x + 220

    def model_dump(self, by_alias: bool = True, **kwargs: any) -> dict:
        """Ensure stages are properly arranged when pipeline definition is stringified."""
        self._auto_arrange_stages()

        # after arranging we can model dump our stages which are not dumped implicitly
        self.stages_data = [stage.model_dump(by_alias=by_alias) for stage in self.stages]
        self.error_stage_data = self.error_stage.model_dump(by_alias=by_alias) if self.error_stage else None
        self.stats_aggregator_stage_data = (
            self.stats_aggregator_stage.model_dump(by_alias=by_alias) if self.stats_aggregator_stage else None
        )
        self.start_event_stages_data = [self.start_event.model_dump(by_alias=by_alias)] if self.start_event else list()
        self.stop_event_stages_data = [self.stop_event.model_dump(by_alias=by_alias)] if self.stop_event else list()
        self.test_origin_stages_data = self.test_origin.model_dump(by_alias=by_alias) if self.test_origin else None

        # metadata labels need to be an empty list for the UI
        if self.metadata is None:
            self.metadata = dict()
        self.metadata.update(dict(labels=list()))

        return super().model_dump(by_alias=by_alias, **kwargs)


class StreamsetsFlowUsage(BaseModel):
    """The Model for StreamsetsFlowUsage."""

    last_update_time: str | None = Field(repr=True)
    access_count: int | None = Field(repr=True)
    last_access_time: str | None = Field(repr=False)
    last_updated_at: str | None = Field(repr=False)
    last_updater_id: str | None = Field(repr=False)
    last_accessed_at: str | None = Field(repr=False)
    last_accessor_id: str | None = Field(repr=False)


class StreamsetsFlowROV(BaseModel):
    """The Model for StreamsetsFlowROV."""

    mode: int | None = Field(repr=True)
    member_roles: dict | None = Field(repr=False)
    collaborator_ids: dict | None = Field(repr=False)


class StreamsetsFlowMetadata(BaseModel):
    """The Model for StreamsetsFlowMetadata."""

    name: str = Field(repr=True)
    description: str = Field(repr=True)
    flow_id: str = Field(
        alias="asset_id",
        repr=True,
    )
    owner_id: str | None = Field(repr=False, default=None)
    created: int | None = Field(repr=False, default=None)
    created_at: str | None = Field(repr=False, default=None)
    tags: list | None = Field(repr=False, default=None)
    asset_attributes: list | None = Field(repr=False, default=None)
    asset_state: str | None = Field(repr=False, default=None)
    asset_type: str | None = Field(repr=False, default=None)
    catalog_id: str | None = Field(repr=False, default=None)
    rating: int | None = Field(repr=False, default=None)
    total_ratings: int | None = Field(repr=False, default=None)
    size: int | None = Field(repr=False, default=None)
    project_id: str | None = Field(repr=False, default=None)
    rov: StreamsetsFlowROV | None = Field(repr=False, default=None)
    usage: StreamsetsFlowUsage | None = Field(repr=False, default=None)
    version: float | None = Field(repr=False, default=None)
    create_time: str | None = Field(repr=False, default=None)
    is_linked_with_sub_container: bool | None = Field(repr=False, default=None)
    asset_category: str | None = Field(repr=False, default=None)
    sandbox_id: str | None = Field(repr=False, default=None)
    creator_id: str | None = Field(repr=False, default=None)

    _expose: bool = PrivateAttr(default=False)


@Flow.register("streamsets")
class StreamsetsFlow(Flow):
    """The Model for StreamsetsFlow."""

    metadata: StreamsetsFlowMetadata = Field(repr=True)

    connection_ids: list[str] = Field(default=[], alias="connection_ids", repr=False)
    engine_version: str = Field(alias="engine_version", repr=True)
    environment_id: str = Field(alias="environment_id", repr=False)
    fragments_ids: list[str] = Field(default=[], alias="fragment_ids", repr=False)

    EXPOSED_DATA_PATH: ClassVar[dict] = {"entity.streamsets_flow": {}}

    def __init__(self, project: "Project", pipeline_definition: dict | None = None, **flow_json: dict) -> None:
        """The __init__ of the StreamsetsFlow class.

        Args:
            project: The Project object.
            pipeline_definition: Pipeline definition for the flow
            flow_json: The JSON for the flow.
        """
        super().__init__(**flow_json)
        self._project = project

        # attributes to be used for lazy-loading properties
        self.__pipeline_definition_object = (
            PipelineDefinition(flow=self, **pipeline_definition) if pipeline_definition else None
        )
        self._environment = None
        self._engine = None

    @staticmethod
    def _create(
        project: "Project",
        name: str,
        environment: Environment | None = None,
        description: str = "",
        flow_type: str = "streamsets",
    ) -> "StreamsetsFlow":
        if flow_type not in SUPPORTED_FLOWS:
            raise TypeError("Flow type not supported.")

        # create a flow in the project
        # default to the latest engine version if an environment is not passed.
        payload = json.dumps(
            {
                "name": name,
                "description": description,
                "environment_id": environment.environment_id if environment else "",
                "engine_version": environment.engine_version
                if environment
                else project._platform.available_engine_versions[0],
            }
        )
        params = {"project_id": project.metadata.guid}
        create_flow_response = project._platform._streamsets_flow_api.create_streamsets_flow(
            params=params, data=payload
        )
        flow = StreamsetsFlow(project=project, **create_flow_response.json()["flow"])

        # create a pipeline definition from the template and add it in.
        flow.pipeline_definition = create_pipeline_definition_for_new_flow(
            flow_executor_id=flow.executor_id,
            iam_id=flow.metadata.usage.last_updater_id,
        )

        # update the flow to ensure it has a pipeline definition
        project.update_flow(flow)

        return flow

    def _update(self) -> requests.Response:
        """Update the Flow class with provided data.

        Args:
            project: The Project object.
            pipeline_definition: Pipeline definition for the flow
            data: The JSON for the flow.
        """
        params = {"project_id": self._project.metadata.guid}
        payload = {
            "name": self.name,
            "connection_ids": self.connection_ids if self.connection_ids else [],
            "description": self.description,
            "environment_id": self.environment_id,
            "engine_version": self.engine_version,
            "fragment_ids": self.fragments_ids if self.fragments_ids else [],
            "pipeline_definition": self.pipeline_definition,
        }

        update_flow_response = self._project._platform._streamsets_flow_api.update_streamsets_flow(
            params=params, flow_id=self.flow_id, data=json.dumps(payload)
        )
        update_flow_json = update_flow_response.json()

        self.__init__(
            project=self._project,
            pipeline_definition=update_flow_json["pipeline_definition"],
            **update_flow_json["flow"],
        )

        return update_flow_response

    def _delete(self) -> requests.Response:
        params = {"project_id": self._project.metadata.guid}
        return self._project._platform._streamsets_flow_api.delete_streamsets_flow(params=params, flow_id=self.flow_id)

    def _duplicate(self, name: str, description: str = "") -> "StreamsetsFlow":
        """Duplicate this StreamsetsFlow with a new name and description.

        Args:
            name: The name for the duplicated flow.
            description: The description for the duplicated flow.

        Returns:
            The newly created duplicate flow.
        """
        payload = json.dumps({"name": name, "description": description})
        params = {"project_id": self._project.metadata.guid}

        response = self._project._platform._streamsets_flow_api.duplicate_streamsets_flow(
            params=params, data=payload, flow_id=self.flow_id
        )
        response_json = response.json()

        flow = StreamsetsFlow(
            project=self._project,
            pipeline_definition=response_json["pipeline_definition"],
            **response_json["flow"],
            flow_type="streamsets",
        )

        return flow

    @property
    def name(self) -> str:
        """Returns name of the flow."""
        return self.metadata.name

    @name.setter
    def name(self, value: str) -> None:
        """Sets name of the flow."""
        self.metadata.name = value

    @property
    def description(self) -> str:
        """Returns description of the flow."""
        return self.metadata.description

    @description.setter
    def description(self, value: str) -> None:
        """Sets description of the flow."""
        self.metadata.description = value

    @property
    def flow_id(self) -> str:
        """Returns id of the flow."""
        return self.metadata.flow_id

    @cached_property
    def executor_id(self) -> str:
        """Get a flow's id in an executor."""
        return f"testRun__{self.metadata.flow_id}"

    @property
    def _pipeline_definition(self) -> PipelineDefinition:
        if self.__pipeline_definition_object is None:
            # lazy load the pipeline definition
            full_flow_data = self._project._platform._streamsets_flow_api.get_streamsets_flow_by_id(
                params=dict(project_id=self.metadata.project_id), flow_id=self.flow_id
            ).json()
            self.__pipeline_definition_object = PipelineDefinition(flow=self, **full_flow_data["pipeline_definition"])
        return self.__pipeline_definition_object

    @property
    def pipeline_definition(self) -> dict:
        """A flow's definition in an engine."""
        return self._pipeline_definition.model_dump(by_alias=True)

    @pipeline_definition.setter
    def pipeline_definition(self, val: dict) -> None:
        self.__pipeline_definition_object = PipelineDefinition(flow=self, **val)

    @property
    def environment(self) -> Environment:
        """The environment this flow belongs to."""
        if self._environment is None and self.environment_id:
            self._environment = self._project.environments.get(**dict(environment_id=self.environment_id))
        return self._environment

    @environment.setter
    def environment(self, val: Environment) -> None:
        """Set the environment of the flow."""
        # unset the _engine attribute for consistency
        self._engine = None

        self._environment = val
        self.environment_id = self._environment.environment_id
        # having an environment, we will always have an associated engine version
        self.engine_version = self._environment.engine_version

    @property
    def engine(self) -> Engine | None:
        """The engine in which this flow will be run."""
        if self._engine is None and self.environment is not None and len(self.environment.engines) > 0:
            self._engine = self.environment.engines[0]
        return self._engine

    @property
    def configuration(self) -> Configuration:
        """The configuration of a flow."""
        return self._pipeline_definition.configuration

    @property
    def stages(self) -> list[Stage | StageWithPredicates]:
        """The stages in the flow."""
        return self._pipeline_definition.stages

    @property
    def error_stage(self) -> Stage | None:
        """The error stage of the flow."""
        return self._pipeline_definition.error_stage

    @property
    def stats_aggregator_stage(self) -> Stage | None:
        """The stats aggregator stage of the flow."""
        return self._pipeline_definition.stats_aggregator_stage

    @property
    def start_event(self) -> Stage | None:
        """The start event stage of the flow."""
        return self._pipeline_definition.start_event

    @property
    def stop_event(self) -> Stage | None:
        """The stop event stage of the flow."""
        return self._pipeline_definition.stop_event

    @property
    def test_origin(self) -> Stage | None:
        """The test origin stage of the flow."""
        return self._pipeline_definition.test_origin

    def add_stage(
        self, label: str | None = None, name: str | None = None, type: str | None = None, library: str | None = None
    ) -> Stage | StageWithPredicates:
        """Add a stage to the flow.

        Args:
            label: Label of the stage to add.
            name: Name of the stage to add.
            type: Type of the stage to add.
            library: Library that stage to add belongs to.
        """
        return self._pipeline_definition.add_stage(label=label, name=name, type=type, library=library)

    def remove_stage(self, stage: Stage) -> None:
        """Removes a stage from the flow.

        Args:
            stage: The stage to remove.
        """
        self._pipeline_definition.remove_stage(stage=stage)

    def set_error_stage(self, label: str | None = None, name: str | None = None, library: str | None = None) -> Stage:
        """Set the error stage of a flow."""
        return self._pipeline_definition.set_error_stage(label=label, name=name, library=library)

    def set_stats_aggregator_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the stats aggregator stage of a flow."""
        return self._pipeline_definition.set_stats_aggregator_stage(label=label, name=name, library=library)

    def set_start_event_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the start event stage of a flow."""
        return self._pipeline_definition.set_start_event_stage(label=label, name=name, library=library)

    def set_stop_event_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the stop event stage of a flow."""
        return self._pipeline_definition.set_stop_event_stage(label=label, name=name, library=library)

    def set_test_origin_stage(
        self, label: str | None = None, name: str | None = None, library: str | None = None
    ) -> Stage:
        """Set the test origin stage of a flow."""
        return self._pipeline_definition.set_test_origin_stage(label=label, name=name, library=library)

    def validate(self) -> list["FlowValidationError"]:
        """Validates a flow.

        Returns:
            A `list` of `FlowValidationError` containing issues.
        """
        if not self.environment:
            raise ValueError("Cannot validate a flow that is not associated with an environment")
        if not self.engine:
            raise ValueError("Cannot validate a flow in an environment without an engine.")

        # create pipeline in sdc if it doesn't exist
        try:
            response = self.engine.api_client.get_pipeline_by_id(pipeline_id=self.executor_id)
        except requests.exceptions.HTTPError:
            response = self.engine.api_client.create_pipeline(
                pipeline_title=self.executor_id,
                params=dict(description="description", draft=False, autoGeneratePipelineId=False),
            )

        # use that engine pipeline's uuid and pipeline id, help's us run on different engines if required.
        pipeline_definition = deepcopy(self.pipeline_definition)
        pipeline_definition["uuid"] = response.json()["uuid"]
        pipeline_definition["pipelineId"] = response.json()["pipelineId"]

        # do the validation
        validation_result = self.engine.api_client.validate_pipeline(self.executor_id)
        previewer_id = validation_result.json()["previewerId"]
        self.engine.api_client.get_pipeline_preview_status(self.executor_id, previewer_id)
        self.engine.api_client.get_pipeline_preview(self.executor_id, previewer_id)

        flow_data_issues = self.engine.api_client.get_pipeline_validation_status(
            self.executor_id, data=pipeline_definition
        ).json()["issues"]

        issues = []
        for k, v in flow_data_issues["stageIssues"].items():
            for i in v:
                validation = FlowValidationError(type="stageIssues", **i)
                issues.append(validation)
        for i in flow_data_issues["pipelineIssues"]:
            validation = FlowValidationError(type="pipelineIssues", **i)
            issues.append(validation)

        return issues

    @override
    def model_dump(self, by_alias: bool = True, **kwargs: any) -> dict:
        data = super().model_dump(by_alias=by_alias, **kwargs)
        data["pipeline_definition"] = self.pipeline_definition
        return data


class StreamsetsFlowPayloadExtender(PayloadExtender):
    """Streamsets flow extender setup also `streamsets_env_id`.

    :meta: private
    """

    @override
    def extend(self, payload: dict[str, Any], flow: Flow) -> dict[str, Any]:
        payload["asset_ref"] = flow.flow_id
        payload.setdefault("configuration", {})["streamsets_env_id"] = flow.environment_id
        return payload


class StreamsetsFlows(CollectionModel):
    """Collection of StreamsetsFlows."""

    def __init__(self, project: "Project") -> None:
        """The __init__ of the StreamsetsFlow class.

        Args:
            project: The Project object.
        """
        super().__init__(project)
        self._project = project
        self.unique_id = "metadata.flow_id"

    def _request_parameters(self) -> list:
        return ["flow_id", "project_id"]

    def __len__(self) -> int:
        """The len of the StreamsetsFlows class."""
        query_params = {
            "project_id": self._project.metadata.guid,
        }
        res = self._project._platform._streamsets_flow_api.get_streamsets_flows(params=query_params)
        res_json = res.json()
        return res_json["total_count"]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        if "start" in request_params:
            parsed_url = urlparse(request_params["start"]["href"])
            params = parse_qs(parsed_url.query)
            request_params["start"] = params.get("start", [None])[0]

        request_params_defaults = {
            "project_id": self._project.metadata.guid,
        }
        request_params_unioned = request_params_defaults
        request_params_unioned.update(request_params)

        if "flow_id" in request_params:
            response_json = self._project._platform._streamsets_flow_api.get_streamsets_flow_by_id(
                params={k: v for k, v in request_params_unioned.items() if v is not None},
                flow_id=request_params["flow_id"],
            ).json()
            response = {"streamsets_flows": [response_json["flow"]]}

        else:
            response = self._project._platform._streamsets_flow_api.get_streamsets_flows(
                params={k: v for k, v in request_params_unioned.items() if v is not None}
            ).json()

        return CollectionModelResults(
            response,
            StreamsetsFlow,
            "next",
            "start",
            "streamsets_flows",
            {"project": self._project, "flow_type": "streamsets"},
        )


class StreamsetsConnectionMetadata(BaseModel):
    """Streamsets Connection Metadata object."""

    asset_category: str | None = Field(default=None, description="Category of the asset", repr=False)
    asset_id: str = Field(description="Asset identifier", repr=True)
    asset_type: str | None = Field(default=None, description="Type of the asset", repr=False)
    create_time: str | None = Field(default=None, description="Creation time", repr=False)
    creator_id: str | None = Field(default=None, description="Asset creator identifier", repr=False)
    project_id: str | None = Field(default=None, description="Project identifier", repr=False)
    model_config = ConfigDict(frozen=True)


class StreamsetsConnection(BaseModel):
    """Streamsets Connection object."""

    EXPOSED_DATA_PATH: ClassVar[dict] = {"entity": {}}
    metadata: StreamsetsConnectionMetadata = Field(repr=False)
    datasource_type: str | None = Field(default=None, description="Datasource type", repr=False)
    name: str | None = Field(default=None, description="Connection name", repr=True)
    properties: dict | None = Field(default=None, repr=False)

    def __init__(self, platform: "Platform" = None, project: "Project" = None, **connection_json: dict) -> None:
        """The __init__ of the StreamsetsConnection class.

        Args:
            connection_json: The JSON for the StreamsetsConnection.
            platform: The Platform object.
            project: The Project object.
        """
        super().__init__(**connection_json)
        self._platform = platform
        self._project = project


class FlowValidationError(BaseModel):
    """Streamsets validation error object."""

    type: str = Field(repr=True)
    instanceName: str | None = Field(default=None, repr=True)
    level: str = Field(repr=False)
    message: str = Field(repr=False)
    errorCode: str = Field(repr=False)
    humanReadableMessage: str = Field(repr=True)
    technicalMessage: str = Field(repr=False)



from ibm_watsonx_data_integration.services.datastage.api.datastage_flow_api import PipelineJson
from typing import Literal


class FlowRunner:
    # def __init__(self, config: Config):
    #     self.project_id = config._get_project_id()
    #     self.catalog_id = None
    #     self.flow_id = None
    #     self.flow_service = config._get_environment()._get_flow_service()
    #     self.job_service = config._get_environment()._get_job_service()
    #     self.env = (
    #         "cpd"
    #         if isinstance(config._get_environment(), CloudPakForDataEnvironment)
    #         else "cloud"
    #     )

    def get_flows(
        self,
        project_id: str | None = None,
        # add in other types of ids?
        entity_name: str | None = None,
    ):
        project_id = project_id or self.project_id
        if not project_id:
            raise ValueError("Project id must be provided")
        response = self.flow_service.list_datastage_flows(project_id=project_id, entity_name=entity_name)
        return response

    def get_flow(self, flow_id: str, project_id: str | None = None):
        project_id = project_id or self.project_id
        if not project_id:
            raise ValueError("Project id must be provided")

        if not flow_id:
            raise ValueError("Flow id must be provided")

        response = self.flow_service.get_datastage_flows(project_id=project_id, data_intg_flow_id=flow_id)
        return response

    def create_flow(
        self,
        flow_name: str,
        flow_json: dict,
        project_id: str | None = None,
        catalog_id: str | None = None,
    ):
        if not flow_name:
            raise ValueError("Flow name must be provided")
        project_id = project_id or self.project_id
        catalog_id = catalog_id or self.catalog_id
        if project_id:
            response = self.flow_service.create_datastage_flows(
                data_intg_flow_name=flow_name,
                pipeline_flows=flow_json,
                project_id=project_id,
            )
            return response
        elif catalog_id:
            response = self.flow_service.create_datastage_flows(
                data_intg_flow_name=flow_name,
                pipeline_flows=flow_json,
                catalog_id=catalog_id,
            )
            return response
        else:
            raise ValueError("Project id or catalog id must be provided")

    def update_flow(
        self,
        flow_name: str,
        flow_json: dict,
        flow_id: str | None = None,
        project_id: str | None = None,
        catalog_id: str | None = None,
    ):
        if not flow_name:
            raise ValueError("Flow name must be provided")
        flow_id = flow_id
        project_id = project_id or self.project_id
        catalog_id = catalog_id or self.catalog_id
        if not flow_id:
            raise ValueError("Create a new flow or provide a flow id")
        if project_id:
            response = self.flow_service.update_datastage_flows(
                data_intg_flow_id=flow_id,
                data_intg_flow_name=flow_name,
                pipeline_flows=flow_json,
                project_id=project_id,
            )
            return response
        elif catalog_id:
            response = self.flow_service.update_datastage_flows(
                data_intg_flow_id=flow_id,
                data_intg_flow_name=flow_name,
                pipeline_flows=flow_json,
                catalog_id=catalog_id,
            )
            return response
        else:
            raise ValueError("Project id or catalog id must be provided")

    def compile_flow(
        self,
        flow_id: str,
        project_id: str | None = None,
        catalog_id: str | None = None,
    ):
        project_id = project_id or self.project_id
        catalog_id = catalog_id or self.catalog_id
        if not flow_id:
            raise ValueError("Create a new flow or provide a flow id")
        if project_id:
            response = self.flow_service.compile_datastage_flows(data_intg_flow_id=flow_id, project_id=project_id)
            return response
        elif catalog_id:
            response = self.flow_service.compile_datastage_flows(data_intg_flow_id=flow_id, catalog_id=catalog_id)
            return response
        else:
            raise ValueError("Project id or catalog id must be provided")

    def get_compile_status(
        self,
        flow_id: str | None = None,
        project_id: str | None = None,
        catalog_id: str | None = None,
    ):
        flow_id = flow_id or self.flow_id
        project_id = project_id or self.project_id
        catalog_id = catalog_id or self.catalog_id
        if not flow_id:
            raise ValueError("Create a new flow or provide a flow id")
        if project_id:
            response = self.flow_service.get_flow_compile_status(data_intg_flow_id=flow_id, project_id=project_id)
            return response
        elif catalog_id:
            response = self.flow_service.get_flow_compile_status(data_intg_flow_id=flow_id, catalog_id=catalog_id)
            return response
        else:
            raise ValueError("Project id or catalog id must be provided")

    def get_compile_info(
        self,
        flow_id: str,
        project_id: str | None = None,
        catalog_id: str | None = None,
    ):
        project_id = project_id or self.project_id
        catalog_id = catalog_id or self.catalog_id
        if not flow_id:
            raise ValueError("Create a new flow or provide a flow id")
        if project_id:
            response = self.flow_service.datastage_flows_compile_info(data_intg_flow_id=flow_id, project_id=project_id)
            return response
        elif catalog_id:
            response = self.flow_service.datastage_flows_compile_info(data_intg_flow_id=flow_id, catalog_id=catalog_id)
            return response
        else:
            raise ValueError("Project id or catalog id must be provided")

    # def get_jobs(self, project_id: str | None = None, asset_ref: str | None = None):
    #     project_id = project_id or self.project_id
    #     # include space_id?

    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     response = self.job_service.jobs_list(
    #         project_id=project_id, asset_ref=asset_ref
    #     )
    #     return response

    # def get_job(self, project_id: str | None = None, job_id: str | None = None):
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     if not job_id:
    #         raise ValueError("Please create a job or provide a job id")

    #     response = self.job_service.jobs_get(job_id=job_id, project_id=project_id)
    #     return response

    # def create_job(
    #     self,
    #     job_name: str,
    #     flow_id: str,
    #     project_id: str | None = None,
    #     paramsets: list = None,
    #     runtime: Runtime = None,
    #     schedule: Schedule = None,
    #     job_body: JobPostBodyJob = None,
    # ):
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     if not flow_id:
    #         raise ValueError("Flow id must be provided")

    #     retention_policy = JobPostBodyJobRetentionPolicy()
    #     configuration = JobPostBodyConfiguration()
    #     schedule_info = JobPostBodyJobScheduleInfo()
    #     schedule_cron = None
    #     if runtime:
    #         if runtime.runtime_settings:
    #             retention_policy = runtime.runtime_settings._get_retention_policy()
    #             configuration = runtime.runtime_settings._get_configuration(
    #                 project_id, self.env
    #             )
    #     if schedule:
    #         schedule_info = JobPostBodyJobScheduleInfo(
    #             repeat=schedule.repeat, start_on=schedule.start, end_on=schedule.end
    #         )
    #         schedule_cron = schedule.schedule

    #     if not job_body:
    #         job = JobPostBodyJob(
    #             name=job_name,
    #             configuration=configuration,
    #             asset_ref=flow_id,
    #             parameter_sets=paramsets,
    #             retention_policy=retention_policy,
    #             schedule_info=schedule_info,
    #             schedule=schedule_cron,
    #         )
    #     else:
    #         job = job_body

    #     response = self.job_service.jobs_create(job=job, project_id=project_id)
    #     return response

    # def update_job(
    #     self,
    #     job_json: dict,
    #     job_id: str,
    #     project_id: str | None = None,
    #     runtime: Runtime = None,
    #     paramsets: list = None,
    #     schedule: Schedule = None,
    # ):
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     if not job_id:
    #         raise ValueError("Please create a job or provide a job id")

    #     configuration = job_json["entity"]["job"]["configuration"]
    #     job = job_json["entity"]["job"]

    #     patches = []

    #     if paramsets:
    #         job["parameter_sets"] = paramsets

    #     if runtime:
    #         if runtime.runtime_settings:
    #             if runtime.runtime_settings.env:
    #                 configuration["env_id"] = (
    #                     runtime.runtime_settings.env.value + "-" + project_id
    #                 )
    #             if runtime.runtime_settings.warn_limit:
    #                 configuration["flow_limits"] = {
    #                     "warn_limit": runtime.runtime_settings.warn_limit
    #                 }
    #             job["configuration"] = configuration
    #             if runtime.runtime_settings.days:
    #                 job["retention_policy"] = {"days": runtime.runtime_settings.days}
    #             if runtime.runtime_settings.amount:
    #                 job["retention_policy"] = {
    #                     "amount": runtime.runtime_settings.amount
    #                 }
    #     if schedule:
    #         if schedule.schedule:
    #             job["schedule"] = schedule.schedule
    #         schedule_info = {}
    #         if schedule.start:
    #             schedule_info["startOn"] = schedule.start
    #         if schedule.end:
    #             schedule_info["endOn"] = schedule.end
    #         if schedule.repeat:
    #             schedule_info["repeat"] = schedule.repeat
    #         job["schedule_info"] = schedule_info

    #     patches.append(
    #         JSONJobPatchModelItem(op="replace", path="/entity/job", value=job)
    #     )

    #     response = self.job_service.jobs_update(
    #         job_id=job_id, body=patches, project_id=project_id
    #     )
    #     return response

    # def run_job(self, job_id: str, project_id: str | None = None):
    #     if not job_id:
    #         raise ValueError("Please create a job or provide a job id")
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     response = self.job_service.job_runs_create(
    #         project_id=project_id,
    #         job_id=job_id,
    #         job_run=JobRunPostBodyJobRunConfiguration(),
    #     )
    #     return response

    # def get_job_run(
    #     self,
    #     job_id: str,
    #     run_id: str,
    #     project_id: str | None = None,
    # ):
    #     if not run_id:
    #         raise ValueError("Please create a job run or provide a run id")
    #     if not job_id:
    #         raise ValueError("Please create a job or provide a job id")
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     response = self.job_service.job_runs_get(
    #         job_id=job_id, run_id=run_id, project_id=project_id
    #     )
    #     return response

    # def get_job_log(
    #     self,
    #     job_id: str,
    #     run_id: str,
    #     project_id: str | None = None,
    # ):
    #     if not run_id:
    #         raise ValueError("Please create a job run or provide a run id")
    #     if not job_id:
    #         raise ValueError("Please create a job or provide a job id")
    #     project_id = project_id or self.project_id
    #     if not project_id:
    #         raise ValueError("Project id must be provided")

    #     response = self.job_service.job_runs_logs(
    #         job_id=job_id, run_id=run_id, project_id=project_id
    #     )
    #     return response

    def create_or_replace_flow(
        self,
        flow_name: str,
        flow_json: dict,
        duplicate_assets: Literal["Skip", "Rename", "Overwrite"],
    ):
        flow_exist_response = self.get_flows(entity_name=flow_name)
        flow_exists = False
        flow_id = None
        for flow in flow_exist_response.result["data_flows"]:
            if flow["metadata"]["name"] == flow_name:
                flow_exists = True
                flow_id = flow["metadata"]["asset_id"]

        pipeline_json = PipelineJson.from_dict(flow_json)
        if flow_exists:
            if duplicate_assets == "Overwrite":
                response = self.update_flow(flow_name=flow_name, flow_json=pipeline_json, flow_id=flow_id)
            if duplicate_assets == "Rename":
                new_name_exists = True
                new_name = None
                i = 0
                while new_name_exists:
                    i += 1
                    new_name = flow_name + "_" + str(i)
                    new_name_exist_response = self.get_flows(entity_name=new_name)
                    found_new_name = False
                    for flow in new_name_exist_response.result["data_flows"]:
                        if flow["metadata"]["name"] == new_name:
                            found_new_name = True
                    new_name_exists = found_new_name
                response = self.create_flow(flow_name=new_name, flow_json=pipeline_json)
            if duplicate_assets == "Skip":
                response = self.get_flow(flow_id=flow_id)
        else:
            response = self.create_flow(flow_name=flow_name, flow_json=pipeline_json)

        return response

    def delete_flow(self, flow_name: str):
        flows_res = self.flow_service.list_datastage_flows(project_id=self.project_id)

        def get_flow_id():
            for flow in flows_res.result["data_flows"]:
                if flow["metadata"]["name"] == flow_name:
                    return flow["metadata"]["asset_id"]

        flow_id = get_flow_id()
        if flow_id is None:
            raise ValueError(f"Error in deleting flow: Flow with name {flow_name} not found")

        self.flow_service.delete_datastage_flows(
            id=[flow_id],
            project_id=self.project_id,
        )
